# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Impurity content defined by ratios, with one species constrained by Z_eff."""
import dataclasses
from typing import Annotated, Literal, Mapping
import chex
import jax
import pydantic
from torax._src import array_typing
from torax._src.torax_pydantic import torax_pydantic
import typing_extensions

# pylint: disable=invalid-name


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams:
  """Runtime parameters for ElectronDensityRatiosZeff."""

  n_e_ratios: Mapping[str, array_typing.FloatVectorCell | None]
  n_e_ratios_face: Mapping[str, array_typing.FloatVectorFace | None]
  unknown_species: str = dataclasses.field(metadata={'static': True})
  Z_override: array_typing.FloatScalar | None = None
  A_override: array_typing.FloatVectorCell | None = None
  A_override_face: array_typing.FloatVectorFace | None = None


class ElectronDensityRatiosZeff(torax_pydantic.BaseModelFrozen):
  """Impurity content defined by ratios, with one species constrained by Z_eff."""

  # Exactly one species must have a None ratio to be constrained by Z_eff.
  species: Mapping[str, torax_pydantic.NonNegativeTimeVaryingArray | None]
  Z_override: torax_pydantic.TimeVaryingScalar | None = None
  A_override: torax_pydantic.TimeVaryingScalar | None = None
  impurity_mode: Annotated[
      Literal['n_e_ratios_Z_eff'], torax_pydantic.JAX_STATIC
  ] = 'n_e_ratios_Z_eff'

  def build_runtime_params(self, t: chex.Numeric) -> RuntimeParams:
    unknown_species = next(
        (symbol for symbol, ratio in self.species.items() if ratio is None),
        None,
    )
    # The validator ensures unknown_species is not None but add an extra check.
    assert unknown_species is not None
    return RuntimeParams(
        n_e_ratios={
            symbol: ratio.get_value(t) if ratio is not None else None
            for symbol, ratio in self.species.items()
        },
        n_e_ratios_face={
            symbol: (
                ratio.get_value(t, grid_type='face')
                if ratio is not None
                else None
            )
            for symbol, ratio in self.species.items()
        },
        unknown_species=unknown_species,
        Z_override=self.Z_override.get_value(t) if self.Z_override else None,
        A_override=self.A_override.get_value(t) if self.A_override else None,
    )

  @pydantic.model_validator(mode='after')
  def _validate_one_none(self) -> typing_extensions.Self:
    if not self.species:
      raise ValueError('The species dictionary cannot be empty.')
    none_count = sum(v is None for v in self.species.values())
    if none_count != 1:
      raise ValueError(
          'Exactly one impurity must have a `None` ratio to be'
          ' constrained by Z_eff.'
      )
    return self



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for torax.plotting.plotruns_lib."""

import os

from absl.testing import absltest
from absl.testing import parameterized
from torax._src.config import config_loader
from torax._src.plotting import plotruns_lib
from torax._src.test_utils import paths


class PlotrunsLibTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    test_data_dir = paths.test_data_dir()
    self.test_data_path = os.path.join(
        test_data_dir, "test_iterhybrid_rampup.nc"
    )

  def test_data_loading(self):
    plotruns_lib.load_data(self.test_data_path)

  @parameterized.parameters([
      "default_plot_config",
      "global_params_plot_config",
      "simple_plot_config",
      "sources_plot_config",
  ])
  def test_plot_config_smoke_test(self, config_name: str):
    config_path = config_loader.torax_path().joinpath(
        "plotting", "configs", config_name + ".py"
    )
    assert config_path.is_file(), f"Path {config_path} is not a file."
    plot_config = config_loader.import_module(config_path)["PLOT_CONFIG"]
    plotruns_lib.plot_run(plot_config, self.test_data_path)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reference values used throughout TORAX unit tests."""
import dataclasses
import json
from typing import Callable, Final, Literal, Mapping

import immutabledict
import jax
import numpy as np
# set_jax_precision() is called in torax __init__.py, needed for tests and
# reference generation
import torax  # pylint: disable=unused-import
from torax._src.config import build_runtime_params
from torax._src.config import config_loader
from torax._src.config import runtime_params_slice
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.geometry import geometry_provider as geometry_provider_lib
from torax._src.torax_pydantic import model_config

# pylint: disable=invalid-name

JSON_FILENAME: Literal['references.json'] = 'references.json'


def _load_all_references():
  """Loads all reference data from the JSON file."""
  json_path = config_loader.torax_path() / '_src' / 'test_utils' / JSON_FILENAME
  try:
    with open(json_path, 'r') as f:
      return json.load(f)
  except FileNotFoundError:
    raise FileNotFoundError(
        f'Reference data file not found at {json_path}.'
    ) from None


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class References:
  """Collection of reference values useful for unit tests."""

  config: model_config.ToraxConfig
  psi: cell_variable.CellVariable
  psi_face_grad: np.ndarray
  psidot: np.ndarray
  j_total: np.ndarray
  q: np.ndarray
  s: np.ndarray

  @property
  def geometry_provider(self) -> geometry_provider_lib.GeometryProvider:
    return self.config.geometry.build_provider

  def get_runtime_params_and_geo(
      self,
  ) -> tuple[runtime_params_slice.RuntimeParams, geometry.Geometry]:
    t = self.config.numerics.t_initial
    params = build_runtime_params.RuntimeParamsProvider.from_config(
        self.config
    )
    return build_runtime_params.get_consistent_runtime_params_and_geometry(
        t=t,
        runtime_params_provider=params,
        geometry_provider=self.config.geometry.build_provider,
    )


def _build_references_from_case_data(
    case_name: str, torax_config: model_config.ToraxConfig
) -> References:
  """Helper function to build a References object from a config and JSON data."""
  geo = torax_config.geometry.build_provider(torax_config.numerics.t_initial)
  all_references = _load_all_references()
  case_data = all_references[case_name]

  psi_value = np.array(case_data['psi'])
  psi_face_grad = np.array(case_data['psi_face_grad'])
  psi_grad_bc = psi_face_grad[-1]

  psi = cell_variable.CellVariable(
      value=psi_value,
      right_face_grad_constraint=psi_grad_bc,
      dr=geo.drho_norm,
  )

  return References(
      config=torax_config,
      psi=psi,
      psi_face_grad=psi_face_grad,
      psidot=np.array(case_data['psidot']),
      j_total=np.array(case_data['j_total']),
      q=np.array(case_data['q']),
      s=np.array(case_data['s']),
  )


def circular_references() -> References:
  """Reference values for circular geometry."""
  torax_config = model_config.ToraxConfig.from_dict({
      'profile_conditions': {
          'Ip': 15e6,
          'current_profile_nu': 3,
          'n_e_nbar_is_fGW': True,
          'normalize_n_e_to_nbar': True,
          'nbar': 0.85,
          'n_e': {0: {0.0: 1.5, 1.0: 1.0}},
      },
      'numerics': {},
      'plasma_composition': {},
      'geometry': {
          'geometry_type': 'circular',
      },
      'transport': {},
      'solver': {},
      'pedestal': {},
      'sources': {'generic_current': {}},
  })
  return _build_references_from_case_data('circular_references', torax_config)


def chease_references_Ip_from_chease() -> References:
  """Reference values for CHEASE geometry where Ip comes from the file."""
  torax_config = model_config.ToraxConfig.from_dict({
      'profile_conditions': {
          'Ip': 15e6,
          'current_profile_nu': 3,
          'n_e_nbar_is_fGW': True,
          'normalize_n_e_to_nbar': True,
          'nbar': 0.85,
          'n_e': {0: {0.0: 1.5, 1.0: 1.0}},
      },
      'numerics': {},
      'plasma_composition': {},
      'geometry': {
          'geometry_type': 'chease',
          'Ip_from_parameters': False,
      },
      'transport': {},
      'solver': {},
      'pedestal': {},
      'sources': {'generic_current': {}},
  })
  return _build_references_from_case_data(
      'chease_references_Ip_from_chease', torax_config
  )


def chease_references_Ip_from_runtime_params() -> References:
  """Reference values for CHEASE geometry where Ip comes from runtime params."""
  torax_config = model_config.ToraxConfig.from_dict({
      'profile_conditions': {
          'Ip': 15e6,
          'current_profile_nu': 3,
          'n_e_nbar_is_fGW': True,
          'normalize_n_e_to_nbar': True,
          'nbar': 0.85,
          'n_e': {0: {0.0: 1.5, 1.0: 1.0}},
      },
      'numerics': {},
      'plasma_composition': {},
      'geometry': {
          'geometry_type': 'chease',
          'Ip_from_parameters': True,
      },
      'transport': {},
      'solver': {},
      'pedestal': {},
      'sources': {'generic_current': {}},
  })
  return _build_references_from_case_data(
      'chease_references_Ip_from_runtime_params', torax_config
  )


REFERENCES_REGISTRY: Final[Mapping[str, Callable[[], References]]] = (
    immutabledict.immutabledict({
        'circular_references': circular_references,
        'chease_references_Ip_from_chease': chease_references_Ip_from_chease,
        'chease_references_Ip_from_runtime_params': (
            chease_references_Ip_from_runtime_params
        ),
    })
)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Carries out the predictor corrector method for the PDE solution.

Picard iterations to approximate the nonlinear solution. If
runtime_params_slice.solver.use_predictor_corrector is False, reverts to
a standard linear solution.
"""
import functools

import jax
from torax._src import jax_utils
from torax._src import state
from torax._src import xnp
from torax._src.config import runtime_params_slice
from torax._src.fvm import block_1d_coeffs
from torax._src.fvm import calc_coeffs
from torax._src.fvm import cell_variable
from torax._src.fvm import implicit_solve_block
from torax._src.geometry import geometry
from torax._src.sources import source_profiles


@functools.partial(
    jax_utils.jit,
    static_argnames=[
        'coeffs_callback',
    ],
)
def predictor_corrector_method(
    dt: jax.Array,
    runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
    geo_t_plus_dt: geometry.Geometry,
    x_old: tuple[cell_variable.CellVariable, ...],
    x_new_guess: tuple[cell_variable.CellVariable, ...],
    core_profiles_t_plus_dt: state.CoreProfiles,
    coeffs_exp: block_1d_coeffs.Block1DCoeffs,
    explicit_source_profiles: source_profiles.SourceProfiles,
    coeffs_callback: calc_coeffs.CoeffsCallback,
) -> tuple[cell_variable.CellVariable, ...]:
  """Predictor-corrector method.

  Args:
    dt: current timestep
    runtime_params_t_plus_dt: Runtime parameters corresponding to the next time
      step, needed for the implicit PDE coefficients.
    geo_t_plus_dt: Geometry at the next time step.
    x_old: Tuple of CellVariables correspond to the evolving core profiles at
      time t.
    x_new_guess: Tuple of CellVariables corresponding to the initial guess for
      the next time step.
    core_profiles_t_plus_dt: Core profiles at the next time step.
    coeffs_exp: Block1DCoeffs PDE coefficients at beginning of timestep.
    explicit_source_profiles: Precomputed explicit source profiles. These
      profiles were configured to always depend on state and parameters at time
      t during the solver step. They can thus be inputs, since they are not
      recalculated at time t+plus_dt with updated state during the solver
      iterations. For sources that are implicit, their explicit profiles are set
      to all zeros.
    coeffs_callback: coefficient callback function.

  Returns:
    x_new: Solution of evolving core profile state variables
  """
  solver_params = runtime_params_t_plus_dt.solver

  # predictor-corrector loop. Will only be traversed once if not in
  # predictor-corrector mode
  def loop_body(i, x_new_guess):  # pylint: disable=unused-argument
    coeffs_new = coeffs_callback(
        runtime_params_t_plus_dt,
        geo_t_plus_dt,
        core_profiles_t_plus_dt,
        x_new_guess,
        explicit_source_profiles=explicit_source_profiles,
        allow_pereverzev=True,
    )

    return implicit_solve_block.implicit_solve_block(
        dt=dt,
        x_old=x_old,
        x_new_guess=x_new_guess,
        coeffs_old=coeffs_exp,
        coeffs_new=coeffs_new,
        theta_implicit=solver_params.theta_implicit,
        convection_dirichlet_mode=(solver_params.convection_dirichlet_mode),
        convection_neumann_mode=(solver_params.convection_neumann_mode),
    )

  if solver_params.use_predictor_corrector:
    x_new = xnp.fori_loop(
        0,
        runtime_params_t_plus_dt.solver.n_corrector_steps + 1,
        loop_body,
        x_new_guess,
    )
  else:
    x_new = loop_body(0, x_new_guess)
  return x_new



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The `optimizer_solve_block` function.

See function docstring for details.
"""
import functools
from typing import TypeAlias

import jax
from torax._src import jax_utils
from torax._src import physics_models as physics_models_lib
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.core_profiles import convertors
from torax._src.fvm import block_1d_coeffs
from torax._src.fvm import calc_coeffs
from torax._src.fvm import cell_variable
from torax._src.fvm import enums
from torax._src.fvm import fvm_conversions
from torax._src.fvm import residual_and_loss
from torax._src.geometry import geometry
from torax._src.solver import predictor_corrector_method
from torax._src.sources import source_profiles

AuxiliaryOutput: TypeAlias = block_1d_coeffs.AuxiliaryOutput


@functools.partial(
    jax_utils.jit,
    static_argnames=[
        'coeffs_callback',
        'evolving_names',
        'initial_guess_mode',
    ],
)
def optimizer_solve_block(
    dt: jax.Array,
    runtime_params_t: runtime_params_slice.RuntimeParams,
    runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
    geo_t: geometry.Geometry,
    geo_t_plus_dt: geometry.Geometry,
    x_old: tuple[cell_variable.CellVariable, ...],
    core_profiles_t: state.CoreProfiles,
    core_profiles_t_plus_dt: state.CoreProfiles,
    explicit_source_profiles: source_profiles.SourceProfiles,
    physics_models: physics_models_lib.PhysicsModels,
    coeffs_callback: calc_coeffs.CoeffsCallback,
    evolving_names: tuple[str, ...],
    initial_guess_mode: enums.InitialGuessMode,
    maxiter: int,
    tol: float,
) -> tuple[
    tuple[cell_variable.CellVariable, ...],
    state.SolverNumericOutputs,
]:
  # pyformat: disable  # pyformat removes line breaks needed for readability
  """Runs one time step of an optimization-based solver on the equation defined by `coeffs`.

  This solver is relatively generic in that it models diffusion, convection,
  etc. abstractly. The caller must do the problem-specific physics calculations
  to obtain the coefficients for a particular problem.

  This solver uses iterative optimization to minimize the norm of the residual
  between two sides of the equation describing a theta method update.

  Args:
    dt: Discrete time step.
    runtime_params_t: Runtime params for time t (the start time of the step).
      These runtime params can change from step to step without triggering a
      recompilation.
    runtime_params_t_plus_dt: Runtime params for time t + dt.
    geo_t: Geometry object used to initialize auxiliary outputs at time t.
    geo_t_plus_dt: Geometry object used to initialize auxiliary outputs at time
      t + dt.
    x_old: Tuple containing CellVariables for each channel with their values at
      the start of the time step.
    core_profiles_t: Core plasma profiles which contain all available prescribed
      quantities at the start of the time step. This includes evolving boundary
      conditions and prescribed time-dependent profiles that are not being
      evolved by the PDE system.
    core_profiles_t_plus_dt: Core plasma profiles which contain all available
      prescribed quantities at the end of the time step. This includes evolving
      boundary conditions and prescribed time-dependent profiles that are not
      being evolved by the PDE system.
    explicit_source_profiles: Pre-calculated sources implemented as explicit
      sources in the PDE.
    physics_models: Physics models used for the calculations.
    coeffs_callback: Calculates diffusion, convection etc. coefficients given a
      core_profiles. Repeatedly called by the iterative optimizer.
    evolving_names: The names of variables within the core profiles that should
      evolve.
    initial_guess_mode: Chooses the initial_guess for the iterative method,
      either x_old or linear step. When taking the linear step, it is also
      recommended to use Pereverzev-Corrigan terms if the transport use
      pereverzev terms for linear solver. Is only applied in the nonlinear
      solver for the optional initial guess from the linear solver.
    maxiter: See docstring of `jaxopt.LBFGS`.
    tol: See docstring of `jaxopt.LBFGS`.

  Returns:
    x_new: Tuple, with x_new[i] giving channel i of x at the next time step
    solver_numeric_outputs: SolverNumericOutputs. Info about iterations and
      errors
  """
  # pyformat: enable

  coeffs_old = coeffs_callback(
      runtime_params_t,
      geo_t,
      core_profiles_t,
      x_old,
      explicit_source_profiles=explicit_source_profiles,
      explicit_call=True,
  )

  match initial_guess_mode:
    # LINEAR initial guess will provide the initial guess using the predictor-
    # corrector method if use_predictor_corrector=True in the solver runtime
    # params
    case enums.InitialGuessMode.LINEAR:
      # returns transport coefficients with additional pereverzev terms
      # if set by runtime_params, needed if stiff transport models (e.g. qlknn)
      # are used.
      coeffs_exp_linear = coeffs_callback(
          runtime_params_t,
          geo_t,
          core_profiles_t,
          x_old,
          explicit_source_profiles=explicit_source_profiles,
          allow_pereverzev=True,
          explicit_call=True,
      )
      # See linear_theta_method.py for comments on the predictor_corrector API
      x_new_guess = convertors.core_profiles_to_solver_x_tuple(
          core_profiles_t_plus_dt, evolving_names
      )
      init_x_new = predictor_corrector_method.predictor_corrector_method(
          dt=dt,
          runtime_params_t_plus_dt=runtime_params_t_plus_dt,
          geo_t_plus_dt=geo_t_plus_dt,
          x_old=x_old,
          x_new_guess=x_new_guess,
          core_profiles_t_plus_dt=core_profiles_t_plus_dt,
          coeffs_exp=coeffs_exp_linear,
          coeffs_callback=coeffs_callback,
          explicit_source_profiles=explicit_source_profiles,
      )
      init_x_new_vec = fvm_conversions.cell_variable_tuple_to_vec(init_x_new)
    case enums.InitialGuessMode.X_OLD:
      init_x_new_vec = fvm_conversions.cell_variable_tuple_to_vec(x_old)
    case _:
      raise ValueError(
          f'Unknown option for first guess in iterations: {initial_guess_mode}'
      )

  solver_numeric_outputs = state.SolverNumericOutputs()

  # Advance jaxopt_solver by one timestep
  (
      x_new_vec,
      final_loss,
      solver_numeric_outputs.inner_solver_iterations,
  ) = residual_and_loss.jaxopt_solver(
      dt=dt,
      runtime_params_t_plus_dt=runtime_params_t_plus_dt,
      geo_t_plus_dt=geo_t_plus_dt,
      x_old=x_old,
      init_x_new_vec=init_x_new_vec,
      core_profiles_t_plus_dt=core_profiles_t_plus_dt,
      explicit_source_profiles=explicit_source_profiles,
      physics_models=physics_models,
      coeffs_old=coeffs_old,
      evolving_names=evolving_names,
      maxiter=maxiter,
      tol=tol,
  )

  # Create updated CellVariable instances based on core_profiles_t_plus_dt which
  # has updated boundary conditions and prescribed profiles.
  x_new = fvm_conversions.vec_to_cell_variable_tuple(
      x_new_vec, core_profiles_t_plus_dt, evolving_names
  )

  # Tell the caller whether or not x_new successfully reduces the loss below
  # the tolerance by providing an extra output, error.
  solver_numeric_outputs.solver_error_state = jax.lax.cond(
      final_loss > tol,
      lambda: 1,  # Called when True
      lambda: 0,  # Called when False
  )

  return x_new, solver_numeric_outputs



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class and utils for TGLF-based models."""
import dataclasses

import jax
from jax import numpy as jnp
from torax._src import array_typing
from torax._src import constants
from torax._src import state
from torax._src.geometry import geometry
from torax._src.physics import psi_calculations
from torax._src.transport_model import quasilinear_transport_model


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams(quasilinear_transport_model.RuntimeParams):
  """Shared parameters for TGLF-based models."""

  #  Left here as a placeholder for future TGLF settings

  pass


# pylint: disable=invalid-name
@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class TGLFInputs(quasilinear_transport_model.QuasilinearInputs):
  r"""Dimensionless inputs to TGLF-based models.

  See https://gacode.io/tglf/tglf_table.html for definitions.

  Attributes:
    Ti_over_Te: Ratio of ion temperature to electron temperature.
    r_minor: Flux surface centroid minor radius.
    dr_major: Gradient of the flux surface centroid major radius with respect to
      the minor radius (:math:`dr_{major}/dr_{minor}`).
    q: The safety factor.
    q_prime: Magnetic shear, defined as :math:`\frac{q^2 a^2 s}{r^2}`.
    nu_ee: The electron-electron collision frequency.
    kappa: Plasma elongation.
    kappa_shear: Shear in elongation, defined as
      :math:`\frac{r}{\kappa} \frac{d\kappa}{dr}`.
    delta: Plasma triangularity.
    delta_shear: Shear in triangularity, defined as :math:`r\frac{d\delta}{dr}`.
    beta_e: Electron pressure normalized by TGLF's :math:`B_\mathrm{unit}`.
    Zeff: Effective charge.
  """

  Ti_over_Te: array_typing.FloatVectorFace
  r_minor: array_typing.FloatVectorFace
  dr_major: array_typing.FloatVectorFace
  q: array_typing.FloatVectorFace
  q_prime: array_typing.FloatVectorFace
  nu_ee: array_typing.FloatVectorFace
  kappa: array_typing.FloatVectorFace
  kappa_shear: array_typing.FloatVectorFace
  delta: array_typing.FloatVectorFace
  delta_shear: array_typing.FloatVectorFace
  beta_e: array_typing.FloatVectorFace
  Zeff: array_typing.FloatVectorFace

  # Also define all the TGLF notations for the variables
  @property
  def TAUS_2(self) -> array_typing.FloatVectorFace:
    return self.Ti_over_Te

  @property
  def DRMAJDX_LOC(self) -> array_typing.FloatVectorFace:
    return self.dr_major

  @property
  def Q_LOC(self) -> array_typing.FloatVectorFace:
    return self.q

  @property
  def Q_PRIME_LOC(self) -> array_typing.FloatVectorFace:
    return self.q_prime

  @property
  def XNUE(self) -> array_typing.FloatVectorFace:
    return self.nu_ee

  @property
  def KAPPA_LOC(self) -> array_typing.FloatVectorFace:
    return self.kappa

  @property
  def S_KAPPA_LOC(self) -> array_typing.FloatVectorFace:
    return self.kappa_shear

  @property
  def DELTA_LOC(self) -> array_typing.FloatVectorFace:
    return self.delta

  @property
  def S_DELTA_LOC(self) -> array_typing.FloatVectorFace:
    return self.delta_shear

  @property
  def BETAE(self) -> array_typing.FloatVectorFace:
    return self.beta_e

  @property
  def ZEFF(self) -> array_typing.FloatVectorFace:
    return self.Zeff

  @property
  def RLNS_1(self) -> array_typing.FloatVectorFace:
    return self.lref_over_lne

  @property
  def RLTS_1(self) -> array_typing.FloatVectorFace:
    return self.lref_over_lte

  @property
  def RLTS_2(self) -> array_typing.FloatVectorFace:
    return self.lref_over_lti

  @property
  def RMIN_LOC(self) -> array_typing.FloatVectorFace:
    return self.r_minor

  @property
  def VEXB_SHEAR(self) -> array_typing.FloatVectorFace:
    # TODO(b/381199010): Replace with real values once rotation added to TORAX
    return jnp.zeros_like(self.r_minor)


class TGLFBasedTransportModel(
    quasilinear_transport_model.QuasilinearTransportModel
):
  """Base class for TGLF-based transport models."""

  def _prepare_tglf_inputs(
      self,
      transport: RuntimeParams,  # pylint: disable=unused-argument
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> TGLFInputs:
    """Construct a TGLFInputs object from the TORAX state.

    Normalisation and coordinate conventions:
      TGLF values are normalised with respect to:
        - Minor radius at the LCFS (a) in m
        - B_unit in T
          (defined at https://gacode.io/geometry.html#effective-field)
        - The ion sound speed in m/s
          (defined at https://gacode.io/cgyro.html#id10)
      The radial coordinate is the minor radius of the flux surface (r) in m.
      Psi_TGLF is Psi_TORAX / 2π.

    Shortcomings:
      Currently, geometry parameters are taken directly from the global
      profiles. For complete accuracy, a Miller (or s-α) geometry fit routine
      should instead be implemented to calculate *local* geometry parameters.

      At present, only a subset of TGLF inputs have been implemented.

      This model is under construction.

    References:
      The main points of reference for this implementation are:
        - The TGLF documentation (https://gacode.io/tglf.html)
        - General GACode documentation (https://gacode.io/input_gacode.html)
        - The CGYRO documentation (https://gacode.io/cgyro.html)

      If something isn't documented in TGLF, it will use the same definitions as
      CGYRO and other GACode libraries. The implementation in this function uses
      SI units and definitions throughout unless specified otherwise; be careful
      when comparing with the TGLF docs, as they use a combination of SI, keV,
      1e19m^-3, and Gaussian CGS.

    Args:
      transport: Runtime parameters for the transport model.
      geo: Geometric parameters of the tokamak.
      core_profiles: Core plasma profiles (e.g., temperatures, densities, q).

    Returns:
      A `TGLFInputs` dataclass containing dimensionless inputs required by
      TGLF-based models.
    """
    T_e = core_profiles.T_e.face_value() * constants.CONSTANTS.keV2J
    n_e = core_profiles.n_e.face_value()

    # Reference values used for TGLF-specific normalisation
    # - 'a' in TGLF means the minor radius at the LCFS
    # - 'r' in TGLF means the flux surface centroid minor radius. Gradients are
    #   taken w.r.t. r
    #   https://gacode.io/tglf/tglf_list.html#rmin-loc
    # - B_unit = 1/r d(psi_tor/2π)/dr = q/2πr dpsi/dr, noting that psi_tor is
    #   psi/2π
    #   https://gacode.io/geometry.html#effective-field
    #   https://gacode.io/cgyro.html#faq
    # - c_s (ion sound speed)
    #   https://gacode.io/cgyro.html#id10
    m_D_amu = constants.ION_PROPERTIES_DICT['D'].A  # Mass of deuterium [amu]
    m_D = m_D_amu * constants.CONSTANTS.m_amu  # Mass of deuterium [kg]
    c_s = (T_e / m_D) ** 0.5  # T_e [J], m_D [kg], gives c_s in [m/s]
    a = geo.a_minor  # Device minor radius at LCFS [m]
    r = geo.r_mid_face  # Flux surface centroid minor radius [m]
    B_unit = (
        core_profiles.q_face
        / (2 * jnp.pi * r)  # Note: psi_TGLF is psi_TORAX/2π
        * jnp.gradient(core_profiles.psi.face_value(), r)
    )

    # Temperature ratio
    Ti_over_Te = core_profiles.T_i.face_value() / core_profiles.T_e.face_value()

    # Dimensionless gradients
    normalized_log_gradients = quasilinear_transport_model.NormalizedLogarithmicGradients.from_profiles(
        core_profiles=core_profiles,
        radial_coordinate=geo.r_mid,  # On the cell grid
        reference_length=a,
    )

    # Dimensionless electron-electron collision frequency = nu_ee / (c_s/a)
    # https://gacode.io/tglf/tglf_list.html#xnue
    # https://gacode.io/cgyro/cgyro_list.html#nu-ee
    # - In the TGLF docs, XNUE is labelled as electron-ion collision
    #   frequency. It is believed that it should actually be the electron-
    #   electron collision frequency, see
    #   https://pyrokinetics.readthedocs.io/en/latest/user_guide/collisions.html#tglf
    # - In the TGLF docs, nu_ee is shown in CGS units. Below is the SI version.
    # - TGLF uses a slightly different calculation of log_Lambda to those in
    #   Wesson, namely
    #       log_Lambda = (
    #         24.0 - 0.5 * jnp.log(n_e_1e19 * 1e13) + jnp.log(T_e_keV * 1e3)
    #       )
    #   https://github.com/gafusion/gacode/blob/740bb2dc811bc1ad38be65ea4b87330995931305/cgyro/src/cgyro_make_profiles.F90#L145
    #   This is different to Wesson by about ~0.5%. Below, we use the TLGF
    #   version, but with n_e [m^-3] and T_e [J].
    log_Lambda = 74.2 - 0.5 * jnp.log(n_e) + jnp.log(T_e)

    # ν_ee = (sqrt(2) n_e q_e^4 lnΛ) / (16 π ε_0^2 m_e^0.5 T_e^1.5)
    # Compute via the log for stability
    log_nu_ee = (
        0.5 * jnp.log(2)
        + jnp.log(n_e)
        + 4 * jnp.log(constants.CONSTANTS.qe)
        + jnp.log(log_Lambda)
        - jnp.log(16)
        - jnp.log(jnp.pi)
        - 2 * jnp.log(constants.CONSTANTS.epsilon0)
        - 0.5 * jnp.log(constants.CONSTANTS.me)
        - 1.5 * jnp.log(T_e)
    )

    normalized_nu_ee = jnp.exp(log_nu_ee) / (c_s / a)

    # Dimensionless safety factor shear
    # https://gacode.io/tglf/tglf_list.html#tglf-q-prime-loc
    # - TGLF definition is q^2 a^2 s / r^2
    #   where s = r/q dq/dr
    #   (https://gacode.io/cgyro/cgyro_list.html#s)
    q_prime = (
        psi_calculations.calc_s_rmid(geo, core_profiles.psi)
        * core_profiles.q_face**2
        * a**2
        / r**2
    )

    # Electron beta
    # https://gacode.io/tglf/tglf_list.html#tglf-betae
    # https://gacode.io/cgyro.html#faq
    # https://gacode.io/cgyro/cgyro_list.html#betae-unit
    # - In the TGLF docs, beta_e equation shown in CGS units, this is the SI
    #   version
    beta_e = 2 * constants.CONSTANTS.mu0 * n_e * T_e / B_unit**2

    # Major radius shear = drmaj/drmin, where 'rmaj' is the flux surface
    # centroid major radius and 'rmin' the flux surface centroid minor radius
    # https://gacode.io/tglf/tglf_list.html#tglf-drmajdx-loc
    r_major = (geo.R_in_face + geo.R_out_face) / 2
    dr_major = jnp.gradient(r_major, r)

    # Dimensionless elongation shear = r/kappa dkappa/dr
    # https://gacode.io/tglf/tglf_list.html#tglf-s-kappa-loc
    kappa = geo.elongation_face
    kappa_shear = r / kappa * jnp.gradient(kappa, r)

    # Dimensionless triangularity shear = r ddelta/dr
    # https://gacode.io/tglf/tglf_list.html#tglf-s-delta-loc
    delta_shear = r * jnp.gradient(geo.delta_face, r)

    # Gyrobohm diffusivity
    # https://gacode.io/tglf/tglf_table.html#id7
    # https://gacode.io/cgyro.html#id10
    # - TGLF uses the same normalisation as CGYRO.
    # - The extra c^2 in the docs comes from CGS Gaussian units when
    #   calculating \rho_s
    chiGB = quasilinear_transport_model.calculate_chiGB(
        reference_temperature=core_profiles.T_e.face_value(),  # [keV]
        reference_magnetic_field=B_unit,
        reference_mass=m_D_amu,  # [amu]
        reference_length=a,
    )

    return TGLFInputs(
        # From QuasilinearInputs
        chiGB=chiGB,
        Rmin=geo.a_minor,
        Rmaj=geo.R_major,
        lref_over_lti=normalized_log_gradients.lref_over_lti,
        lref_over_lte=normalized_log_gradients.lref_over_lte,
        lref_over_lne=normalized_log_gradients.lref_over_lne,
        lref_over_lni0=normalized_log_gradients.lref_over_lni0,
        lref_over_lni1=normalized_log_gradients.lref_over_lni1,
        # From TGLFInputs
        Ti_over_Te=Ti_over_Te,
        r_minor=r / a,
        dr_major=dr_major,
        q=core_profiles.q_face,
        q_prime=q_prime,
        nu_ee=normalized_nu_ee,
        kappa=kappa,
        kappa_shear=kappa_shear,
        delta=geo.delta_face,
        delta_shear=delta_shear,
        beta_e=beta_e,
        Zeff=core_profiles.Z_eff_face,
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Polymorphic JAX/NumPy module."""

import contextlib
import functools
import threading
from typing import Any, Callable, TYPE_CHECKING, TypeVar

from absl import logging as native_logging
import jax
import jax.numpy as jnp
import numpy as np
from torax._src import jax_utils

# Export all symbols from jax.numpy API for type checkers (including editors).
# pylint: disable=wildcard-import
# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top
if TYPE_CHECKING:
  from jax.numpy import *
# pylint: enable=wildcard-import
# pylint: enable=g-bad-import-order
# pylint: enable=g-import-not-at-top


T = TypeVar('T')
BooleanNumeric = Any  # A bool, or a Boolean array.

# Create thread-local storage.
thread_context = threading.local()


@contextlib.contextmanager
def _jit_context():
  """A context manager that sets a thread-local flag to indicate JAX context."""
  original_is_jax = getattr(thread_context, 'is_jax', None)
  thread_context.is_jax = True
  try:
    yield
  finally:
    if original_is_jax is None:
      del thread_context.is_jax
    else:
      thread_context.is_jax = original_is_jax


def jit(*args, **kwargs):
  """JAX jit wrapper that sets the thread-local flag to indicate JAX context."""
  func = args[0]

  @functools.wraps(func)
  def wrapped_func(*func_args, **func_kwargs):
    with _jit_context():
      return func(*func_args, **func_kwargs)

  # if TORAX_EXPERIMENTAL_COMPILE is not set use default `False` and do not JIT.
  if jax_utils.env_bool('TORAX_EXPERIMENTAL_COMPILE', False):
    return jax.jit(wrapped_func, *args[1:], **kwargs)
  else:
    return func


def py_while(
    cond_fun: Callable[list[T], BooleanNumeric],
    body_fun: Callable[list[T], T],
    init_val: T,
) -> T:
  """Pure Python implementation of jax.lax.while_loop.

  This gives us a way to write code that could easily be changed to be
  Jax-compatible in the future (if we want to compute its gradient or
  compile it, etc.) without having to pay the high compile time cost
  of jax.lax.while_loop.

  Args:
    cond_fun: function of type ``a -> Bool``.
    body_fun: function of type ``a -> a``.
    init_val: value of type ``a``, a type that can be a scalar, array, or any
      pytree (nested Python tuple/list/dict) thereof, representing the initial
      loop carry value.

  Returns:
    The output from the final iteration of body_fun, of type ``a``.

  .. _Haskell-like type signature: https://wiki.haskell.org/Type_signature
  """

  val = init_val
  while cond_fun(val):
    val = body_fun(val)
  return val


def while_loop(
    cond_fun: Callable[list[T], BooleanNumeric],
    body_fun: Callable[list[T], T],
    init_val: T,
):
  is_jax = getattr(thread_context, 'is_jax', False)
  if is_jax:
    return jax.lax.while_loop(cond_fun, body_fun, init_val)
  else:
    return py_while(cond_fun, body_fun, init_val)


# pylint: disable=g-bare-generic
def py_cond(
    cond_val: bool,
    true_fun: Callable,
    false_fun: Callable,
    *operands,
) -> Any:
  """Pure Python implementation of jax.lax.cond.

  This gives us a way to write code that could easily be changed to be
  Jax-compatible in the future, if we want to expand the scope of the jit
  compilation.

  Args:
    cond_val: The condition.
    true_fun: Function to be called if cond==True.
    false_fun: Function to be called if cond==False.
    *operands: The operands to be passed to the functions.

  Returns:
    The output from either true_fun or false_fun.
  """
  if cond_val:
    return true_fun(*operands)
  else:
    return false_fun(*operands)


def cond(
    cond_val: bool,
    true_fun: Callable[..., Any],  # pytype: disable=invalid-annotation
    false_fun: Callable[..., Any],  # pytype: disable=invalid-annotation
    *operands,
) -> Any:
  is_jax = getattr(thread_context, 'is_jax', False)
  if is_jax:
    return jax.lax.cond(cond_val, true_fun, false_fun, *operands)
  else:
    return py_cond(cond_val, true_fun, false_fun, *operands)


def py_fori_loop(
    lower: int, upper: int, body_fun: Callable[[int, T], T], init_val: T  # pytype: disable=invalid-annotation
) -> T:
  """Pure Python implementation of jax.lax.fori_loop.

  This gives us a way to write code that could easily be changed to be
  Jax-compatible in the future, if we want to expand the scope of the jit
  compilation.

  Args:
    lower: lower integer of loop
    upper: upper integer of loop. upper<=lower will produce no iterations.
    body_fun: function of type ``a -> a``.
    init_val: value of type ``a``, a type that can be a scalar, array, or any
      pytree (nested Python tuple/list/dict) thereof, representing the initial
      loop carry value.

  Returns:
    The output from the final iteration of body_fun, of type ``a``.

  .. _Haskell-like type signature: https://wiki.haskell.org/Type_signature
  """
  val = init_val
  for i in range(lower, upper):
    val = body_fun(i, val)
  return val


def fori_loop(
    lower: int,
    upper: int,
    body_fun: Callable[..., Any],  # pytype: disable=invalid-annotation
    init_val: Any,
):
  is_jax = getattr(thread_context, 'is_jax', False)
  if is_jax:
    return jax.lax.fori_loop(lower, upper, body_fun, init_val)
  else:
    return py_fori_loop(lower, upper, body_fun, init_val)


def logging(msg: Any, *args: Any, **kwargs: Any):
  """Logging wrapper that works under xnp.jit.

  This function makes use of jax.debug.print when running under xnp.jit,
  otherwise it uses native logging.

  Note, jax.debug.print makes use of jax.debug.callback behind the scenes which:
  - does not guarantee execution.
  https://docs.jax.dev/en/latest/external-callbacks.html#flavors-of-callback
  - will cause cache misses when using the persistent cache.
  https://docs.jax.dev/en/latest/persistent_compilation_cache.html#pitfalls

  Therefore advise using for debug logging only if running under xnp.jit.

  Args:
    msg: message to log.
    *args: arguments to be passed to the logging function.
    **kwargs: keyword arguments to be passed to the logging function.
  """
  is_jax = getattr(thread_context, 'is_jax', False)
  if is_jax:
    jax.debug.print(msg, *args, **kwargs)
  else:
    native_logging.info(msg, *args, **kwargs)


def _get_current_lib():
  """Determines whether to use JAX or NumPy."""
  is_jax = getattr(thread_context, 'is_jax', False)
  if is_jax:
    return jnp
  else:
    return np


def __getattr__(name):  # pylint: disable=invalid-name
  """Returns the corresponding function from the current library (jnp or np)."""
  current_lib = _get_current_lib()
  try:
    return getattr(current_lib, name)
  except AttributeError as exc:
    raise AttributeError(f"Module 'xnp' has no attribute '{name}'") from exc


def __dir__():  # pylint: disable=invalid-name
  """Provides a list of potential attributes for code completion."""
  common_attributes = set(dir(jnp)) | set(dir(np))

  return sorted(list(common_attributes))



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base pydantic config and model for sawtooth redistribution."""

import abc

import chex
from torax._src import array_typing
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.mhd.sawtooth import runtime_params as sawtooth_runtime_params
from torax._src.torax_pydantic import torax_pydantic


class RedistributionModel(abc.ABC):
  """Abstract base class for sawtooth redistribution models."""

  @abc.abstractmethod
  def __call__(
      self,
      rho_norm_q1: array_typing.FloatScalar,
      runtime_params: runtime_params_slice.RuntimeParams,
      geo: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
  ) -> state.CoreProfiles:
    """Returns a redistributed core_profiles if sawtooth has been triggered."""

  @abc.abstractmethod
  def __hash__(self) -> int:
    """Returns a hash of the redistribution model.

    Should be implemented to support jax.jit caching.
    """

  @abc.abstractmethod
  def __eq__(self, other) -> bool:
    """Equality method to be implemented to support jax.jit caching."""


class RedistributionConfig(torax_pydantic.BaseModelFrozen):
  """Base config for all redistribution models.

  Attributes:
    flattening_factor: The factor by which the profile is flattened.
    Default is near 1.0 but not exactly 1.0, to avoid zero-gradients.
  """

  flattening_factor: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(1.01)
  )

  def build_runtime_params(
      self, t: chex.Numeric
  ) -> sawtooth_runtime_params.RedistributionRuntimeParams:
    return sawtooth_runtime_params.RedistributionRuntimeParams(
        flattening_factor=self.flattening_factor.get_value(t),
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class for handling QLKNN10D models."""

from collections.abc import Mapping
import json
import os
from typing import Any, Callable, Final

import flax.linen as nn
import immutabledict
import jax
import jax.numpy as jnp
import numpy as np
from torax._src import jax_utils
from torax._src.transport_model import base_qlknn_model
from torax._src.transport_model import qualikiz_based_transport_model
import typing_extensions

# Internal import.
# Internal import.


QLKNN10D_NAME: Final[str] = 'qlknn10D'


_ACTIVATION_FNS: Final[Mapping[str, Callable[[jax.Array], jax.Array]]] = (
    immutabledict.immutabledict({
        'relu': nn.relu,
        'tanh': nn.tanh,
        'sigmoid': nn.sigmoid,
        'none': lambda x: x,
    })
)


class MLP(nn.Module):
  hidden_sizes: list[int]
  activations: list[str]

  @nn.compact
  def __call__(self, x: jax.Array) -> jax.Array:
    for act, size in zip(self.activations, self.hidden_sizes):
      x = _ACTIVATION_FNS[act](nn.Dense(size)(x))
    return x


class QuaLiKizNDNN:
  """Jax implementation for QLKNN10D inference."""

  def __init__(
      self,
      model_config: dict[str, Any],
  ):
    self._model_config = model_config

    self._feature_names = model_config.get('feature_names')
    self._target_names = model_config.get('target_names')

    self._feature_prescale_factor = self._load_prescale(
        'prescale_factor', self._feature_names
    )
    self._feature_prescale_bias = self._load_prescale(
        'prescale_bias', self._feature_names
    )
    self._target_prescale_factor = self._load_prescale(
        'prescale_factor', self._target_names
    )
    self._target_prescale_bias = self._load_prescale(
        'prescale_bias', self._target_names
    )

    activations = model_config['hidden_activation'] + [
        model_config['output_activation']
    ]
    hidden_sizes = []
    params = {}
    for i in range(len(activations)):
      weights = np.array(model_config[f'layer{i+1}/weights/Variable:0'])
      params[f'Dense_{i}'] = {
          'bias': np.array(model_config[f'layer{i+1}/biases/Variable:0']),
          'kernel': weights,
      }
      hidden_sizes.append(weights.shape[1])
      del model_config[f'layer{i+1}/biases/Variable:0']
      del model_config[f'layer{i+1}/weights/Variable:0']
    self._params = {'params': params}
    self._model = MLP(hidden_sizes=hidden_sizes, activations=activations)

  def _load_prescale(self, key: str, names: list[str]) -> np.ndarray:
    return np.array([self._model_config[key][k] for k in names])[np.newaxis, :]

  def __call__(
      self,
      inputs: jax.Array,
  ) -> jax.Array:
    """Calculate the outputs given specific inputs."""

    inputs = (
        self._feature_prescale_factor * inputs + self._feature_prescale_bias
    )
    outputs = self._model.apply(self._params, inputs)
    outputs = (
        outputs - self._target_prescale_bias
    ) / self._target_prescale_factor

    return outputs

  @classmethod
  def from_json(cls, json_file) -> typing_extensions.Self:
    with open(json_file) as file_:
      model_dict = json.load(file_)
    return cls(model_dict)


class QLKNN10D(base_qlknn_model.BaseQLKNNModel):
  """Class holding QLKNN10D networks.

  Attributes:
    path: Path to qlknn-hyper
    net_itgleading: ITG Qi net
    net_itgqediv: ITG Qe/Qi net
    net_temleading: TEM Qe net
    net_temqediv: TEM Qi/Qe net
    net_etgleading: ETG Qe net
    net_temqidiv: TEM Qi/Qe net
    net_tempfediv: Tem pfe/Qe net
    net_etgleading: ITG Qe/Qi net
    net_itgpfediv: ITG pfe/Qi net
  """

  def __init__(self, path: str, name: str):
    # Only one version of qlknn10D, so we ignore the name.
    del name
    super().__init__(path=path, name=QLKNN10D_NAME)
    self.net_itgleading = self._load('efiitg_gb.json')
    self.net_itgqediv = self._load('efeitg_gb_div_efiitg_gb.json')
    self.net_temleading = self._load('efetem_gb.json')
    self.net_temqidiv = self._load('efitem_gb_div_efetem_gb.json')
    self.net_tempfediv = self._load('pfetem_gb_div_efetem_gb.json')
    self.net_etgleading = self._load('efeetg_gb.json')
    self.net_itgpfediv = self._load('pfeitg_gb_div_efiitg_gb.json')

  def _load(self, path) -> QuaLiKizNDNN:
    full_path = os.path.join(self.path, path)
    return QuaLiKizNDNN.from_json(full_path)

  def predict(
      self,
      inputs: jax.Array,
  ) -> base_qlknn_model.ModelOutput:
    """Feed forward through the network and compute fluxes."""

    model_output = {}
    model_output['qi_itg'] = self.net_itgleading(inputs).clip(0)
    model_output['qe_itg'] = self.net_itgqediv(inputs) * model_output['qi_itg']
    model_output['pfe_itg'] = (
        self.net_itgpfediv(inputs) * model_output['qi_itg']
    )
    model_output['qe_tem'] = self.net_temleading(inputs).clip(0)
    model_output['qi_tem'] = self.net_temqidiv(inputs) * model_output['qe_tem']
    model_output['pfe_tem'] = (
        self.net_tempfediv(inputs) * model_output['qe_tem']
    )
    model_output['qe_etg'] = self.net_etgleading(inputs).clip(0)
    return model_output

  def get_model_inputs_from_qualikiz_inputs(
      self, qualikiz_inputs: qualikiz_based_transport_model.QualikizInputs
  ) -> jax.Array:
    """Converts QualikizInputs to model inputs."""
    return jnp.array(
        [
            getattr(qualikiz_inputs, key)
            for key in self.inputs_and_ranges.keys()
        ],
        dtype=jax_utils.get_dtype(),
    ).T

  @property
  def inputs_and_ranges(self) -> base_qlknn_model.InputsAndRanges:
    return {
        'Z_eff_face': {'min': 1.0, 'max': 3.0},
        'Ati': {'min': 0.0, 'max': 14.0},
        'Ate': {'min': 0.0, 'max': 14.0},
        'Ane': {'min': -5.0, 'max': 6.0},
        'q': {'min': 0.66, 'max': 15.0},
        'smag': {'min': -1.0, 'max': 5.0},
        'x': {'min': 0.09, 'max': 0.99},
        'Ti_Te': {'min': 0.25, 'max': 2.5},
        'log_nu_star_face': {'min': -5.0, 'max': 0.0},
    }



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pydantic model for the neoclassical package."""

import copy
from typing import Any

import pydantic
from torax._src.neoclassical import neoclassical_models
from torax._src.neoclassical import runtime_params
from torax._src.neoclassical.bootstrap_current import sauter as sauter_current
from torax._src.neoclassical.bootstrap_current import zeros as bootstrap_current_zeros
from torax._src.neoclassical.conductivity import sauter as sauter_conductivity
from torax._src.neoclassical.transport import angioni_sauter
from torax._src.neoclassical.transport import zeros as transport_zeros
from torax._src.torax_pydantic import torax_pydantic


class Neoclassical(torax_pydantic.BaseModelFrozen):
  """Config for neoclassical models."""

  bootstrap_current: (
      bootstrap_current_zeros.ZerosModelConfig
      | sauter_current.SauterModelConfig
  ) = pydantic.Field(discriminator="model_name")
  conductivity: sauter_conductivity.SauterModelConfig = (
      torax_pydantic.ValidatedDefault(sauter_conductivity.SauterModelConfig())
  )
  transport: (
      transport_zeros.ZerosModelConfig | angioni_sauter.AngioniSauterModelConfig
  ) = pydantic.Field(discriminator="model_name")

  @pydantic.model_validator(mode="before")
  @classmethod
  def _defaults(cls, data: dict[str, Any]) -> dict[str, Any]:
    configurable_data = copy.deepcopy(data)
    # Set zero models if model not in config dict.
    if "bootstrap_current" not in configurable_data:
      configurable_data["bootstrap_current"] = {"model_name": "zeros"}
    if "transport" not in configurable_data:
      configurable_data["transport"] = {"model_name": "zeros"}
    # Set default model names.
    if "model_name" not in configurable_data["bootstrap_current"]:
      configurable_data["bootstrap_current"]["model_name"] = "sauter"
    if "model_name" not in configurable_data["transport"]:
      configurable_data["transport"]["model_name"] = "angioni_sauter"

    return configurable_data

  def build_runtime_params(self) -> runtime_params.RuntimeParams:
    return runtime_params.RuntimeParams(
        bootstrap_current=self.bootstrap_current.build_runtime_params(),
        conductivity=self.conductivity.build_runtime_params(),
        transport=self.transport.build_runtime_params(),
    )

  def build_models(self) -> neoclassical_models.NeoclassicalModels:
    return neoclassical_models.NeoclassicalModels(
        conductivity=self.conductivity.build_model(),
        bootstrap_current=self.bootstrap_current.build_model(),
        transport=self.transport.build_model(),
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Impurity content defined by ratios of impurity to electron density."""
import dataclasses
from typing import Annotated, Literal, Mapping
import chex
import jax
from jax import numpy as jnp
import jaxtyping as jt
import pydantic
from torax._src import array_typing
from torax._src import constants
from torax._src.torax_pydantic import torax_pydantic
import typing_extensions

# pylint: disable=invalid-name


def calculate_fractions_from_ratios(
    ratios: array_typing.FloatVector,
) -> array_typing.FloatVector:
  """Calculates fractions from ratios, handling the all-zero case."""
  # Ratios can be 1D (n_species,) or 2D (n_species, n_grid).
  # Sum over the species axis.
  total_ratio = jnp.sum(ratios, axis=0)

  is_positive = total_ratio > 0.0

  # Avoid division by zero by replacing zeros in total_ratio with 1.0.
  # The result of this division will be masked out by jnp.where anyway.
  safe_total_ratio = jnp.where(is_positive, total_ratio, 1.0)

  # Calculate fractions where total_ratio is positive.
  calculated_fractions = ratios / safe_total_ratio

  # For the zero impurity case, return uniform fractions to avoid NaNs.
  # The choice is arbitrary as it will be multiplied by zero impurity density.
  num_species = ratios.shape[0]
  # num_species guaranteed to be > 0 since no empty impurity dict is allowed.
  uniform_fractions = jnp.ones_like(ratios) / num_species

  return jnp.where(is_positive, calculated_fractions, uniform_fractions)


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class RuntimeParams:
  """Analogous to ion_mixture.RuntimeParams but for n_e_ratio inputs."""

  n_e_ratios: jt.Float[array_typing.Array, 'ion_symbol rhon']
  n_e_ratios_face: jt.Float[array_typing.Array, 'ion_symbol rhon+1']
  A_avg: array_typing.FloatVectorCell
  A_avg_face: array_typing.FloatVectorFace
  Z_override: array_typing.FloatScalar | None = None

  @property
  def fractions(self) -> array_typing.FloatVector:
    """Returns the impurity fractions calculated from the n_e_ratios."""
    return calculate_fractions_from_ratios(self.n_e_ratios)

  @property
  def fractions_face(self) -> array_typing.FloatVectorFace:
    """Returns the impurity fractions calculated from the n_e_ratios."""
    return calculate_fractions_from_ratios(self.n_e_ratios_face)


class ElectronDensityRatios(torax_pydantic.BaseModelFrozen):
  """Impurity content defined by ratios of impurity to electron density.

  Attributes:
    species: A dictionary of TimeVaryingArray objects, where the keys are ion
      symbols and the values are TimeVaryingArray objects representing the
      ratios of impurity to electron density at the defined rho_norm points.
    Z_override: A TimeVaryingScalar object representing the charge override.
    A_override: A TimeVaryingScalar object representing the A override.
  """

  species: Mapping[str, torax_pydantic.NonNegativeTimeVaryingArray]
  Z_override: torax_pydantic.TimeVaryingScalar | None = None
  A_override: torax_pydantic.TimeVaryingScalar | None = None
  impurity_mode: Annotated[Literal['n_e_ratios'], torax_pydantic.JAX_STATIC] = (
      'n_e_ratios'
  )

  @pydantic.model_validator(mode='after')
  def _validate_species_not_empty(self) -> typing_extensions.Self:
    if not self.species:
      raise ValueError('The species dictionary cannot be empty.')
    return self

  def build_runtime_params(self, t: chex.Numeric) -> RuntimeParams:
    """Creates a RuntimeParams object at a given time."""
    ions = self.species.keys()
    n_e_ratios_arr = jnp.array(
        [ratio.get_value(t) for ratio in self.species.values()]
    )
    n_e_ratios_face_arr = jnp.array([
        ratio.get_value(t, grid_type='face') for ratio in self.species.values()
    ])
    Z_override = None if not self.Z_override else self.Z_override.get_value(t)
    fractions = calculate_fractions_from_ratios(n_e_ratios_arr)
    fractions_face = calculate_fractions_from_ratios(n_e_ratios_face_arr)

    if not self.A_override:
      As = jnp.array([constants.ION_PROPERTIES_DICT[ion].A for ion in ions])
      A_avg = jnp.sum(As[..., jnp.newaxis] * fractions, axis=0)
      A_avg_face = jnp.sum(As[..., jnp.newaxis] * fractions_face, axis=0)
    else:
      A_override = self.A_override.get_value(t)
      A_avg = jnp.ones_like(n_e_ratios_arr[0]) * A_override
      A_avg_face = jnp.ones_like(n_e_ratios_face_arr[0]) * A_override

    return RuntimeParams(
        n_e_ratios=n_e_ratios_arr,
        n_e_ratios_face=n_e_ratios_face_arr,
        A_avg=A_avg,
        Z_override=Z_override,
        A_avg_face=A_avg_face,
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for calculating impurity radiation outputs per impurity species."""

import dataclasses

import jax
from jax import numpy as jnp
import numpy as np
from torax._src import array_typing
from torax._src.config import runtime_params_slice
from torax._src.orchestration import sim_state as sim_state_lib
from torax._src.physics import charge_states
from torax._src.sources import runtime_params as source_runtime_params_lib
from torax._src.sources.impurity_radiation_heat_sink import impurity_radiation_heat_sink
from torax._src.sources.impurity_radiation_heat_sink import impurity_radiation_mavrin_fit
import xarray as xr


RADIATION_OUTPUT_NAME = "radiation_impurity_species"
DENSITY_OUTPUT_NAME = "n_impurity_species"
Z_OUTPUT_NAME = "Z_impurity_species"
IMPURITY_DIM = "impurity_symbol"


# pylint: disable=invalid-name
@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class ImpuritySpeciesOutput:
  """Impurity radiation output for a single species.

  Attributes:
    radiation: The impurity radiation for the species [Wm^-3].
    n_impurity: The impurity density [m^-3].
    Z_impurity: The impurity average charge state.
  """

  radiation: array_typing.FloatVectorCell
  n_impurity: array_typing.FloatVectorCell
  Z_impurity: array_typing.FloatVectorCell


def calculate_impurity_species_output(
    sim_state: sim_state_lib.ToraxSimState,
    runtime_params: runtime_params_slice.RuntimeParams,
) -> dict[str, ImpuritySpeciesOutput]:
  """Calculates the output for each impurity species.

  Args:
    sim_state: The current state of the simulation.
    runtime_params: The runtime parameters for the current simulation slice.

  Returns:
    A dictionary mapping impurity symbols to their `ImpuritySpeciesOutput`.
    If the mavrin impurity radiation heat sink is not enabled then we output
    zeros for the radiation output.
  """
  impurity_species_output: dict[str, ImpuritySpeciesOutput] = {}
  mavrin_active = True
  # If the impurity radiation heat sink is not enabled, return empty dictionary.
  if (
      impurity_radiation_heat_sink.ImpurityRadiationHeatSink.SOURCE_NAME
      not in runtime_params.sources
  ):
    mavrin_active = False
  else:
    runtime_params_impurity = runtime_params.sources[
        impurity_radiation_heat_sink.ImpurityRadiationHeatSink.SOURCE_NAME
    ]
    # If the impurity radiation heat sink is not the mavrin model and in model
    # based mode, return empty dictionary.
    if not (
        isinstance(
            runtime_params_impurity, impurity_radiation_mavrin_fit.RuntimeParams
        )
        and runtime_params_impurity.mode
        == source_runtime_params_lib.Mode.MODEL_BASED
    ):
      mavrin_active = False

  impurity_fractions = sim_state.core_profiles.impurity_fractions
  impurity_names = runtime_params.plasma_composition.impurity_names
  charge_state_info = charge_states.get_average_charge_state(
      ion_symbols=impurity_names,
      T_e=sim_state.core_profiles.T_e.value,
      fractions=jnp.stack(
          [impurity_fractions[symbol] for symbol in impurity_names]
      ),
      Z_override=runtime_params.plasma_composition.impurity.Z_override,
  )

  for i, symbol in enumerate(impurity_names):
    core_profiles = sim_state.core_profiles
    impurity_density_scaling = (
        core_profiles.Z_impurity / charge_state_info.Z_avg
    )
    n_imp = (
        impurity_fractions[symbol]
        * core_profiles.n_impurity.value
        * impurity_density_scaling
    )
    Z_imp = charge_state_info.Z_per_species[i]
    if mavrin_active:
      lz = impurity_radiation_mavrin_fit.calculate_impurity_radiation_single_species(
          core_profiles.T_e.value, symbol
      )
      radiation = n_imp * core_profiles.n_e.value * lz
    else:
      radiation = jnp.zeros_like(n_imp)
    impurity_species_output[symbol] = ImpuritySpeciesOutput(
        radiation=radiation, n_impurity=n_imp, Z_impurity=Z_imp
    )

  return impurity_species_output


def construct_xarray_for_radiation_output(
    impurity_radiation_outputs: dict[str, ImpuritySpeciesOutput],
    times: jax.Array,
    rho_cell_norm: jax.Array,
    time_coord: str,
    rho_cell_norm_coord: str,
) -> dict[str, xr.DataArray]:
  """Constructs an xarray dictionary for impurity radiation output."""

  radiation_data = []
  n_impurity_data = []
  Z_impurity_data = []
  impurity_symbols = []
  xr_dict = {}

  for impurity_symbol in impurity_radiation_outputs:
    radiation_data.append(impurity_radiation_outputs[impurity_symbol].radiation)
    n_impurity_data.append(
        impurity_radiation_outputs[impurity_symbol].n_impurity
    )
    Z_impurity_data.append(
        impurity_radiation_outputs[impurity_symbol].Z_impurity
    )
    impurity_symbols.append(impurity_symbol)
  radiation_data = np.stack(radiation_data, axis=0)
  n_impurity_data = np.stack(n_impurity_data, axis=0)
  Z_impurity_data = np.stack(Z_impurity_data, axis=0)
  xr_dict[RADIATION_OUTPUT_NAME] = xr.DataArray(
      radiation_data,
      dims=[IMPURITY_DIM, time_coord, rho_cell_norm_coord],
      coords={
          IMPURITY_DIM: impurity_symbols,
          time_coord: times,
          rho_cell_norm_coord: rho_cell_norm,
      },
      name=RADIATION_OUTPUT_NAME,
  )
  xr_dict[DENSITY_OUTPUT_NAME] = xr.DataArray(
      n_impurity_data,
      dims=[IMPURITY_DIM, time_coord, rho_cell_norm_coord],
      coords={
          IMPURITY_DIM: impurity_symbols,
          time_coord: times,
          rho_cell_norm_coord: rho_cell_norm,
      },
      name=DENSITY_OUTPUT_NAME,
  )
  xr_dict[Z_OUTPUT_NAME] = xr.DataArray(
      Z_impurity_data,
      dims=[IMPURITY_DIM, time_coord, rho_cell_norm_coord],
      coords={
          IMPURITY_DIM: impurity_symbols,
          time_coord: times,
          rho_cell_norm_coord: rho_cell_norm,
      },
      name=Z_OUTPUT_NAME,
  )
  return xr_dict



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Class for impurity radiation heat sinks.

Model functions are in separate files.
"""

import dataclasses

from torax._src.sources import source as source_lib


@dataclasses.dataclass(kw_only=True, frozen=True, eq=True)
class ImpurityRadiationHeatSink(source_lib.Source):
  """Impurity radiation heat sink for electron heat equation."""

  SOURCE_NAME = "impurity_radiation"
  model_func: source_lib.SourceProfileFunction

  @property
  def source_name(self) -> str:
    return self.SOURCE_NAME

  @property
  def affected_core_profiles(
      self,
  ) -> tuple[source_lib.AffectedCoreProfile, ...]:
    return (source_lib.AffectedCoreProfile.TEMP_EL,)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The NonLinearThetaMethod class."""
import abc
import dataclasses

import jax
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.core_profiles import convertors
from torax._src.fvm import calc_coeffs
from torax._src.fvm import cell_variable
from torax._src.fvm import enums
from torax._src.fvm import newton_raphson_solve_block
from torax._src.fvm import optimizer_solve_block
from torax._src.geometry import geometry
from torax._src.solver import runtime_params
from torax._src.solver import solver
from torax._src.sources import source_profiles


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class OptimizerRuntimeParams(runtime_params.RuntimeParams):
  n_max_iterations: int
  loss_tol: float
  initial_guess_mode: int = dataclasses.field(metadata={'static': True})


@jax.tree_util.register_dataclass
@dataclasses.dataclass(frozen=True)
class NewtonRaphsonRuntimeParams(runtime_params.RuntimeParams):
  maxiter: int
  residual_tol: float
  residual_coarse_tol: float
  delta_reduction_factor: float
  tau_min: float
  initial_guess_mode: int = dataclasses.field(metadata={'static': True})
  log_iterations: bool = dataclasses.field(metadata={'static': True})


class NonlinearThetaMethod(solver.Solver):
  """Time step update using nonlinear solvers and the theta method."""

  def _x_new(
      self,
      dt: jax.Array,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
      core_profiles_t_plus_dt: state.CoreProfiles,
      explicit_source_profiles: source_profiles.SourceProfiles,
      evolving_names: tuple[str, ...],
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      state.SolverNumericOutputs,
  ]:
    """See Solver._x_new docstring."""

    coeffs_callback = calc_coeffs.CoeffsCallback(
        physics_models=self.physics_models,
        evolving_names=evolving_names,
    )
    (
        x_new,
        solver_numeric_outputs,
    ) = self._x_new_helper(
        dt=dt,
        runtime_params_t=runtime_params_t,
        runtime_params_t_plus_dt=runtime_params_t_plus_dt,
        geo_t=geo_t,
        geo_t_plus_dt=geo_t_plus_dt,
        core_profiles_t=core_profiles_t,
        core_profiles_t_plus_dt=core_profiles_t_plus_dt,
        explicit_source_profiles=explicit_source_profiles,
        coeffs_callback=coeffs_callback,
        evolving_names=evolving_names,
    )

    return (
        x_new,
        solver_numeric_outputs,
    )

  @abc.abstractmethod
  def _x_new_helper(
      self,
      dt: jax.Array,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
      core_profiles_t_plus_dt: state.CoreProfiles,
      explicit_source_profiles: source_profiles.SourceProfiles,
      coeffs_callback: calc_coeffs.CoeffsCallback,
      evolving_names: tuple[str, ...],
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      state.SolverNumericOutputs,
  ]:
    """Abstract method for subclasses to implement the specific nonlinear solve.

    This helper method is called by `_x_new` after it has constructed the
    `coeffs_callback`. Subclasses should implement this method to call their
    respective nonlinear solver implementation (e.g., Newton-Raphson or an
    optimizer-based approach).

    Args:
      dt: Time step duration.
      runtime_params_t: Runtime parameters for time t (the start time of the
        step).
      runtime_params_t_plus_dt: Runtime parameters for time t + dt, used for
        implicit calculations in the solver.
      geo_t: Magnetic geometry at time t.
      geo_t_plus_dt: Magnetic geometry at time t + dt.
      core_profiles_t: Core plasma profiles at the beginning of the time step.
      core_profiles_t_plus_dt: Core plasma profiles which contain all available
        prescribed quantities at the end of the time step. This includes
        evolving boundary conditions and prescribed time-dependent profiles that
        are not being evolved by the PDE system.
      explicit_source_profiles: Pre-calculated sources implemented as explicit
        sources in the PDE.
      coeffs_callback: Calculates diffusion, convection etc. coefficients given
        a core_profiles, geometry, runtime_params. Repeatedly called by the
        iterative solvers.
      evolving_names: The names of variables within the core profiles that
        should evolve.

    Returns:
      A tuple containing:
        - The new values of the evolving variables at time t + dt.
        - Solver iteration and error info.
    """
    ...


class OptimizerThetaMethod(NonlinearThetaMethod):
  """Minimize the squared norm of the residual of the theta method equation."""

  def _x_new_helper(
      self,
      dt: jax.Array,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
      core_profiles_t_plus_dt: state.CoreProfiles,
      explicit_source_profiles: source_profiles.SourceProfiles,
      coeffs_callback: calc_coeffs.CoeffsCallback,
      evolving_names: tuple[str, ...],
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      state.SolverNumericOutputs,
  ]:
    """See abstract method docstring in NonlinearThetaMethod."""
    solver_params = runtime_params_t.solver
    assert isinstance(solver_params, OptimizerRuntimeParams)
    (
        x_new,
        solver_numeric_outputs,
    ) = optimizer_solve_block.optimizer_solve_block(
        dt=dt,
        runtime_params_t=runtime_params_t,
        runtime_params_t_plus_dt=runtime_params_t_plus_dt,
        geo_t=geo_t,
        geo_t_plus_dt=geo_t_plus_dt,
        x_old=convertors.core_profiles_to_solver_x_tuple(
            core_profiles_t, evolving_names
        ),
        core_profiles_t=core_profiles_t,
        core_profiles_t_plus_dt=core_profiles_t_plus_dt,
        physics_models=self.physics_models,
        explicit_source_profiles=explicit_source_profiles,
        coeffs_callback=coeffs_callback,
        evolving_names=evolving_names,
        initial_guess_mode=enums.InitialGuessMode(
            solver_params.initial_guess_mode,
        ),
        maxiter=solver_params.n_max_iterations,
        tol=solver_params.loss_tol,
    )
    return (
        x_new,
        solver_numeric_outputs,
    )


class NewtonRaphsonThetaMethod(NonlinearThetaMethod):
  """Nonlinear theta method using Newton Raphson."""

  def _x_new_helper(
      self,
      dt: jax.Array,
      runtime_params_t: runtime_params_slice.RuntimeParams,
      runtime_params_t_plus_dt: runtime_params_slice.RuntimeParams,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
      core_profiles_t_plus_dt: state.CoreProfiles,
      explicit_source_profiles: source_profiles.SourceProfiles,
      coeffs_callback: calc_coeffs.CoeffsCallback,
      evolving_names: tuple[str, ...],
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      state.SolverNumericOutputs,
  ]:
    """See abstract method docstring in NonlinearThetaMethod."""
    solver_params = runtime_params_t.solver
    assert isinstance(solver_params, NewtonRaphsonRuntimeParams)

    (
        x_new,
        solver_numeric_outputs,
    ) = newton_raphson_solve_block.newton_raphson_solve_block(
        dt=dt,
        runtime_params_t=runtime_params_t,
        runtime_params_t_plus_dt=runtime_params_t_plus_dt,
        geo_t=geo_t,
        geo_t_plus_dt=geo_t_plus_dt,
        x_old=convertors.core_profiles_to_solver_x_tuple(
            core_profiles_t, evolving_names
        ),
        core_profiles_t=core_profiles_t,
        core_profiles_t_plus_dt=core_profiles_t_plus_dt,
        explicit_source_profiles=explicit_source_profiles,
        physics_models=self.physics_models,
        coeffs_callback=coeffs_callback,
        evolving_names=evolving_names,
        log_iterations=solver_params.log_iterations,
        initial_guess_mode=enums.InitialGuessMode(
            solver_params.initial_guess_mode
        ),
        maxiter=solver_params.maxiter,
        tol=solver_params.residual_tol,
        coarse_tol=solver_params.residual_coarse_tol,
        delta_reduction_factor=solver_params.delta_reduction_factor,
        tau_min=solver_params.tau_min,
    )
    return (
        x_new,
        solver_numeric_outputs,
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for building source profiles in TORAX."""
import functools

from torax._src import array_typing
from torax._src import jax_utils
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.neoclassical import neoclassical_models as neoclassical_models_lib
from torax._src.neoclassical.bootstrap_current import base as bootstrap_current_base
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.sources import source as source_lib
from torax._src.sources import source_models as source_models_lib
from torax._src.sources import source_profiles
from torax._src.sources.impurity_radiation_heat_sink import impurity_radiation_heat_sink

_FINAL_SOURCES = frozenset(
    [impurity_radiation_heat_sink.ImpurityRadiationHeatSink.SOURCE_NAME]
)


@functools.partial(
    jax_utils.jit,
    static_argnames=[
        'source_models',
        'neoclassical_models',
        'explicit',
    ],
)
def build_source_profiles(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    source_models: source_models_lib.SourceModels,
    neoclassical_models: neoclassical_models_lib.NeoclassicalModels,
    explicit: bool,
    explicit_source_profiles: source_profiles.SourceProfiles | None = None,
    conductivity: conductivity_base.Conductivity | None = None,
) -> source_profiles.SourceProfiles:
  """Builds explicit profiles or the union of explicit and implicit profiles.

  Args:
    runtime_params: Input config for this time step. Can change from time step
      to time step.
    geo: Geometry of the torus.
    core_profiles: Core plasma profiles, either at the start of the time step
      (if explicit) or the live profiles being evolved during the time step (if
      implicit).
    source_models: Functions computing profiles for all TORAX sources/sinks.
    neoclassical_models: Functions computing neoclassical physics.
    explicit: If True, this function will only return profiles for explicit
      sources. If False, then the explicit_source_profiles argument must be
      provided and the returned profiles will be the union of the explicit
      profiles in explicit_source_profiles and the implicit profiles computed
      here. Note that for the special-case sources (bootstrap and qei), the
      explicit argument will be used to determine whether to return a profile or
      all zeros.
    explicit_source_profiles: If explicit is False, this argument must be
      provided. It will be used to compute the union of the explicit profiles
      and the implicit profiles computed here.
    conductivity: Conductivity calculated for this time step. Not provided when
      calculating the explicit profiles.

  Returns:
    SourceProfiles caclulated from the source models. If explicit is True, then
    only explicit profiles will be returned. If explicit is False, then the
    union of the explicit profiles in explicit_source_profiles and the implicit
    profiles computed here will be returned.
  """
  if not explicit and explicit_source_profiles is None:
    raise ValueError(
        '`explicit_source_profiles` must be provided if explicit is False.'
    )

  if explicit:
    qei = source_profiles.QeiInfo.zeros(geo)
    bootstrap_current = bootstrap_current_base.BootstrapCurrent.zeros(geo)
  else:
    qei = source_models.qei_source.get_qei(
        runtime_params=runtime_params,
        geo=geo,
        core_profiles=core_profiles,
    )
    bootstrap_current = (
        neoclassical_models.bootstrap_current.calculate_bootstrap_current(
            runtime_params, geo, core_profiles
        )
    )
  profiles = source_profiles.SourceProfiles(
      bootstrap_current=bootstrap_current,
      qei=qei,
      T_e=explicit_source_profiles.T_e if explicit_source_profiles else {},
      T_i=explicit_source_profiles.T_i if explicit_source_profiles else {},
      n_e=explicit_source_profiles.n_e if explicit_source_profiles else {},
      psi=explicit_source_profiles.psi if explicit_source_profiles else {},
  )
  build_standard_source_profiles(
      calculated_source_profiles=profiles,
      runtime_params=runtime_params,
      geo=geo,
      core_profiles=core_profiles,
      source_models=source_models,
      explicit=explicit,
      conductivity=conductivity,
  )
  return profiles


def build_standard_source_profiles(
    *,
    calculated_source_profiles: source_profiles.SourceProfiles,
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    source_models: source_models_lib.SourceModels,
    explicit: bool = True,
    conductivity: conductivity_base.Conductivity | None = None,
    calculate_anyway: bool = False,
    psi_only: bool = False,
):
  """Updates calculated_source_profiles with standard source profiles."""

  def calculate_source(source_name: str, source: source_lib.Source):
    source_params = runtime_params.sources[source_name]
    if (explicit == source_params.is_explicit) | calculate_anyway:
      value = source.get_value(
          runtime_params,
          geo,
          core_profiles,
          calculated_source_profiles,
          conductivity,
      )
      _update_standard_source_profiles(
          calculated_source_profiles,
          source_name,
          source.affected_core_profiles,
          value,
      )

  # Calculate PSI sources first
  # These are used in the calculation of the ohmic_heat_source so we need to
  # calculate them first.
  for source_name, source in source_models.psi_sources.items():
    calculate_source(source_name, source)
  # The psi sources are used in the initialization of the core profiles
  # (specifically, the psidot, psi and currents), so we provide an option to
  # only calculate the psi sources to avoid extra work.
  if psi_only:
    return

  to_calculate = {}
  # Calculate the standard sources
  # This calculates all the remaining sources that have not been calculated
  # yet and are not final sources.
  # Final sources are sources that depend on the output of other sources, so
  # they are calculated after all other sources have been calculated.
  # The impurity_radiation_heat_sink is one such final source and needs to be
  # calculated after all the heat sources have been calculated.
  for source_name, source in source_models.standard_sources.items():
    if source_name in _FINAL_SOURCES:
      to_calculate[source_name] = source
      continue
    if source_name not in source_models.psi_sources:
      calculate_source(source_name, source)
  for source_name, source in to_calculate.items():
    calculate_source(source_name, source)


def _update_standard_source_profiles(
    calculated_source_profiles: source_profiles.SourceProfiles,
    source_name: str,
    affected_core_profiles: tuple[source_lib.AffectedCoreProfile, ...],
    profile: tuple[array_typing.FloatVectorCell, ...],
):
  """Updates the standard source profiles in calculated_source_profiles.

  Args:
    calculated_source_profiles: The source profiles to update with the newly
      calculated source profiles. Here we are adding to the standard source
      dictionaries so are modifying the calculated_source_profiles in place.
    source_name: The name of the source.
    affected_core_profiles: The core profiles affected by the source.
    profile: The profile of the source.
  """
  for profile, affected_core_profile in zip(
      profile, affected_core_profiles, strict=True
  ):
    match affected_core_profile:
      case source_lib.AffectedCoreProfile.PSI:
        calculated_source_profiles.psi[source_name] = profile
      case source_lib.AffectedCoreProfile.NE:
        calculated_source_profiles.n_e[source_name] = profile
      case source_lib.AffectedCoreProfile.TEMP_ION:
        calculated_source_profiles.T_i[source_name] = profile
      case source_lib.AffectedCoreProfile.TEMP_EL:
        calculated_source_profiles.T_e[source_name] = profile


def build_all_zero_profiles(
    geo: geometry.Geometry,
) -> source_profiles.SourceProfiles:
  """Returns a SourceProfiles object with all zero profiles."""
  return source_profiles.SourceProfiles(
      bootstrap_current=bootstrap_current_base.BootstrapCurrent.zeros(geo),
      qei=source_profiles.QeiInfo.zeros(geo),
  )


def get_all_source_profiles(
    runtime_params: runtime_params_slice.RuntimeParams,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    source_models: source_models_lib.SourceModels,
    neoclassical_models: neoclassical_models_lib.NeoclassicalModels,
    conductivity: conductivity_base.Conductivity,
) -> source_profiles.SourceProfiles:
  """Returns all source profiles for a given time.

  Used e.g. to initialize the source profiles at an initial time step.

  Args:
    runtime_params: Runtime parameters which may change from time step to time
      step without triggering recompilations.
    geo: The geometry of the torus during this time step of the simulation.
    core_profiles: Core profiles that may evolve throughout the course of a
      simulation. These values here are, of course, only the original states.
    source_models: Source models used to compute core source profiles.
    neoclassical_models: Neoclassical models.
    conductivity: Conductivity calculated for this time step.

  Returns:
    Implicit and explicit SourceProfiles from source models based on the core
    profiles from the starting state.
  """
  # Also add in the explicit sources to the initial sources.
  explicit_source_profiles = build_source_profiles(
      runtime_params=runtime_params,
      geo=geo,
      core_profiles=core_profiles,
      source_models=source_models,
      neoclassical_models=neoclassical_models,
      explicit=True,
  )
  return build_source_profiles(
      runtime_params=runtime_params,
      geo=geo,
      core_profiles=core_profiles,
      source_models=source_models,
      neoclassical_models=neoclassical_models,
      explicit=False,
      explicit_source_profiles=explicit_source_profiles,
      conductivity=conductivity,
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Angioni-Sauter neoclassical transport model.

This module implements the neoclassical transport model described in:
C. Angioni and O. Sauter, Phys. Plasmas 7, 1224 (2000).
https://doi.org/10.1063/1.873933

The implementation was facilitated by and verified against the NEOS code:
https://gitlab.epfl.ch/spc/public/neos [O. Sauter et al]
"""

from typing import Annotated, Literal

from jax import numpy as jnp
from torax._src import array_typing
from torax._src import constants
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry as geometry_lib
from torax._src.neoclassical import formulas
from torax._src.neoclassical.transport import base
from torax._src.neoclassical.transport import runtime_params as transport_runtime_params
from torax._src.physics import collisions
from torax._src.torax_pydantic import torax_pydantic
from typing_extensions import override

# pylint: disable=invalid-name


class AngioniSauterModelConfig(base.NeoclassicalTransportModelConfig):
  """Pydantic model for the Angioni-Sauter neoclassical transport model."""

  model_name: Annotated[
      Literal['angioni_sauter'], torax_pydantic.JAX_STATIC
  ] = 'angioni_sauter'

  @override
  def build_model(self) -> 'AngioniSauterModel':
    return AngioniSauterModel()

  @override
  def build_runtime_params(self) -> transport_runtime_params.RuntimeParams:
    return super().build_runtime_params()


class AngioniSauterModel(base.NeoclassicalTransportModel):
  """Implements the Angioni-Sauter neoclassical transport model."""

  @override
  def _call_implementation(
      self,
      runtime_params: runtime_params_slice.RuntimeParams,
      geometry: geometry_lib.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> base.NeoclassicalTransport:
    """Calculates neoclassical transport coefficients."""
    return _calculate_angioni_sauter_transport(
        runtime_params=runtime_params,
        geometry=geometry,
        core_profiles=core_profiles,
    )

  def __hash__(self) -> int:
    return hash(self.__class__.__name__)

  def __eq__(self, other) -> bool:
    return isinstance(other, self.__class__)


def _calculate_angioni_sauter_transport(
    runtime_params: runtime_params_slice.RuntimeParams,
    geometry: geometry_lib.Geometry,
    core_profiles: state.CoreProfiles,
) -> base.NeoclassicalTransport:
  """JIT-compatible implementation of the Angioni-Sauter transport model.

  Args:
    runtime_params: Runtime parameters.
    geometry: Geometry object.
    core_profiles: Core profiles object.

  Returns:
    Neoclassical transport coefficients.

  All internally assigned profiles are on the face grid. The face suffix is
  omitted for brevity.
  """

  del runtime_params  # Unused.

  # --- Step 1: Calculate intermediate physics quantities ---

  # Calculate trapped fractions ft and ftd from paper Eq. (17)
  B2_avg_Bm2_avg = geometry.gm5_face * geometry.gm4_face

  # Use the Sauter model's effective trapped fraction logic
  aa = (1.0 - geometry.epsilon_face) / (1.0 + geometry.epsilon_face)
  epseff = (
      0.67
      * (1.0 - 1.4 * jnp.abs(geometry.delta_face) * geometry.delta_face)
      * geometry.epsilon_face
  )
  ftrap = 1.0 - jnp.sqrt(aa) * (1.0 - epseff) / (1.0 + 2.0 * jnp.sqrt(epseff))

  # Equation (17)
  ftrap_d = 1.0 - (1.0 - ftrap) / B2_avg_Bm2_avg

  # Collisionalities
  log_lambda_ei = collisions.calculate_log_lambda_ei(
      core_profiles.T_e.face_value(), core_profiles.n_e.face_value()
  )
  nu_e_star = formulas.calculate_nu_e_star(
      q=core_profiles.q_face,
      geo=geometry,
      n_e=core_profiles.n_e.face_value(),
      T_e=core_profiles.T_e.face_value(),
      Z_eff=core_profiles.Z_eff_face,
      log_lambda_ei=log_lambda_ei,
  )

  log_lambda_ii = collisions.calculate_log_lambda_ii(
      core_profiles.T_i.face_value(),
      core_profiles.n_i.face_value(),
      core_profiles.Z_i_face,
  )

  # Equation 18c from Sauter PoP 1999
  nu_i_star = (
      4.9e-18
      * core_profiles.q_face
      * geometry.R_major
      * core_profiles.n_i.face_value()
      * core_profiles.Z_i_face**4
      * log_lambda_ii
      / (
          (core_profiles.T_i.face_value() * 1e3) ** 2
          * (geometry.epsilon_face + constants.CONSTANTS.eps) ** 1.5
      )
  )

  # Impurity strength parameter
  # Using single impurity definition: alpha_I = nimpZimp/niZi = Zeff - 1
  alpha_I = core_profiles.Z_eff_face - 1.0

  # --- Step 2: Calculate dimensionless transport matrix K_mn ---
  Kmn_e, Kmn_i = _calculate_Kmn(
      ftrap=ftrap,
      ftrap_d=ftrap_d,
      Z_eff=core_profiles.Z_eff_face,
      B2_avg_Bm2_avg=B2_avg_Bm2_avg,
      nu_e_star=nu_e_star,
      nu_i_star=nu_i_star,
      alpha_I=alpha_I,
  )

  # --- Step 3: Calculate dimensional transport matrix L_mn ---
  Lmn_e, Lmn_i = _calculate_Lmn(
      Kmn_e=Kmn_e,
      Kmn_i=Kmn_i,
      geo=geometry,
      core_profiles=core_profiles,
      epsilon=geometry.epsilon_face,
      nu_e_star=nu_e_star,
      nu_i_star=nu_i_star,
  )

  # --- Step 4: Calculate thermodynamic forces ---
  dpsi_drhon = core_profiles.psi.face_grad() + constants.CONSTANTS.eps
  dlnne_dpsi = (
      core_profiles.n_e.face_grad() / core_profiles.n_e.face_value()
  ) / (dpsi_drhon + constants.CONSTANTS.eps)
  dlnte_dpsi = (
      core_profiles.T_e.face_grad() / core_profiles.T_e.face_value()
  ) / (dpsi_drhon + constants.CONSTANTS.eps)
  dlnni_dpsi = (
      core_profiles.n_i.face_grad() / core_profiles.n_i.face_value()
  ) / (dpsi_drhon + constants.CONSTANTS.eps)
  dlnti_dpsi = (
      core_profiles.T_i.face_grad() / core_profiles.T_i.face_value()
  ) / (dpsi_drhon + constants.CONSTANTS.eps)

  # --- Step 5: Calculate neoclassical fluxes ---
  pe = (
      core_profiles.n_e.face_value()
      * core_profiles.T_e.face_value()
      * constants.CONSTANTS.keV2J
  )
  pi = (
      core_profiles.n_i.face_value()
      * core_profiles.T_i.face_value()
      * constants.CONSTANTS.keV2J
  )
  Rpe = pe / (pe + pi)
  alpha = -Kmn_i[:, 0, 1]
  E_parallel = core_profiles.psidot.face_value() / (
      2 * jnp.pi * geometry.R_major
  )

  # Total electron heat flux Q_e = B_e2 * T_e / (dpsi/drho) (see Angioni Sec 5)
  Be2 = (
      Lmn_e[:, 1, 0] * dlnne_dpsi
      + (Lmn_e[:, 1, 0] + Lmn_e[:, 1, 1]) * dlnte_dpsi
      + (1 - Rpe) / Rpe * Lmn_e[:, 1, 0] * dlnni_dpsi
      + (1 - Rpe) / Rpe * (Lmn_e[:, 1, 0] + alpha * Lmn_e[:, 1, 3]) * dlnti_dpsi
      + Lmn_e[:, 1, 2] * E_parallel / geometry.B_0
  )

  # Total ion heat flux Q_i = B_i2 / T_i * (dpsi/drho) (see Angioni Sec 5)
  Bi2 = (
      alpha * Lmn_e[:, 3, 0] * dlnne_dpsi
      + alpha * (Lmn_e[:, 3, 0] + Lmn_e[:, 3, 1]) * dlnte_dpsi
      + alpha * (1 - Rpe) / Rpe * Lmn_e[:, 3, 0] * dlnni_dpsi
      + alpha * Lmn_e[:, 3, 2] * E_parallel / geometry.B_0
      + (
          Lmn_i[:, 1, 1]
          + (1 - Rpe) / Rpe * alpha**2 / core_profiles.Z_i_face * Lmn_e[:, 3, 3]
      )
      * dlnti_dpsi
  )

  # --- Step 6: Extract transport coefficients --- Angioni Section 5.

  # Q_e = - chi_e * n_e * dT/drho = B_e2 * T_e / dpsi/drho
  # Q_i = - chi_i * n_i * dT/drho = B_i2 * T_i / dpsi/drho
  chi_neo_e = -Be2 / (
      core_profiles.n_e.face_value()
      * dlnte_dpsi
      * (dpsi_drhon / geometry.rho_b) ** 2
      + constants.CONSTANTS.eps
  )
  chi_neo_i = -Bi2 / (
      core_profiles.n_i.face_value()
      * dlnti_dpsi
      * (dpsi_drhon / geometry.rho_b) ** 2
      + constants.CONSTANTS.eps
  )

  # Decomposition of particle flux Be1 = Gamma * dpsi_drho. Page 1232+1233.

  # Diffusive part of particle flux
  # D_e * dn_e/drho  = - L00 *dlog(n_e)/dpsi / dpsi/drho
  D_neo_e = -Lmn_e[:, 0, 0] / (
      core_profiles.n_e.face_value() * (dpsi_drhon / geometry.rho_b) ** 2
      + constants.CONSTANTS.eps
  )

  # Convective part of particle flux, apart from the Ware Pinch term
  # V*n*dpsi/rho = (L00+L01)*dlog(Te)/dpsi + (1-Rpe)/Rpe*L00*dlog(ni)/dpsi +
  # (1-Rpe)/Rpe * (L00+alpha*L03) *dlog(Ti)/dpsi
  V_neo_e = (
      (Lmn_e[:, 0, 0] + Lmn_e[:, 0, 1]) * dlnte_dpsi
      + (1 - Rpe) / Rpe * Lmn_e[:, 0, 0] * dlnni_dpsi
      + (1 - Rpe) / Rpe * (Lmn_e[:, 0, 0] + alpha * Lmn_e[:, 0, 3]) * dlnti_dpsi
  ) / (
      dpsi_drhon / geometry.rho_b * core_profiles.n_e.face_value()
      + constants.CONSTANTS.eps
  )

  # Ware pinch term component of particle convection
  # V_ware*n*dpsi/rho = L02*<E_parallel * B>/<B^2>
  V_neo_ware_e = (
      Lmn_e[:, 0, 2]
      * E_parallel
      / (
          geometry.B_0
          * (dpsi_drhon / geometry.rho_b)
          * core_profiles.n_e.face_value()
          + constants.CONSTANTS.eps
      )
  )

  return base.NeoclassicalTransport(
      chi_neo_i=chi_neo_i,
      chi_neo_e=chi_neo_e,
      D_neo_e=D_neo_e,
      V_neo_e=V_neo_e,
      V_neo_ware_e=V_neo_ware_e,
  )


def _calculate_Kmn(
    ftrap: array_typing.FloatVectorFace,
    ftrap_d: array_typing.FloatVectorFace,
    Z_eff: array_typing.FloatVectorFace,
    B2_avg_Bm2_avg: array_typing.FloatVectorFace,
    nu_e_star: array_typing.FloatVectorFace,
    nu_i_star: array_typing.FloatVectorFace,
    alpha_I: array_typing.FloatVectorFace,
) -> tuple[array_typing.Array, array_typing.Array]:
  """Calculates the dimensionless transport matrices Kmn."""

  # F_mn matrix, Eq. (24)
  F_ftrap = _Fmn_X(ftrap, Z_eff)
  F_ftrap_d = _Fmn_X(ftrap_d, Z_eff)

  # Coefficients from Appendix B.
  a_coeffs, b_coeffs, c_coeffs, d_coeffs, a2, b2, c2, d2 = _coeffs_appendix_B(
      Z_eff
  )

  # Effective collisionalities and geometrical factors
  # Eq. (29)
  nu_e_star_eff = nu_e_star / (1.0 + 7.0 * ftrap**2)
  nu_e_star_eff_sqrt = jnp.sqrt(nu_e_star_eff)
  nu_i_star_eff = nu_i_star / (1.0 + 7.0 * ftrap**2)
  # Eq. (30g)
  FPS = 1.0 - 1.0 / B2_avg_Bm2_avg
  FPS4 = B2_avg_Bm2_avg - 1.0

  # Banana-regime coefficients
  # Eq. 23
  K11e_0 = -0.5 * F_ftrap_d[:, 0, 0]
  K12e_0 = 0.75 * F_ftrap_d[:, 0, 1]
  K22e_0 = -(13.0 / 8.0 + 1 / jnp.sqrt(2.0) / Z_eff) * F_ftrap_d[:, 1, 1]
  K14e_0 = -0.5 * F_ftrap[:, 0, 0]
  K24e_0 = 0.75 * F_ftrap[:, 0, 1]

  # Eq. 30c
  H11_0 = K11e_0
  H12_0 = K12e_0 + 2.5 * K11e_0
  H22_0 = K22e_0 + 5.0 * K12e_0 - 6.25 * K11e_0

  # Eq. 30f
  H41_0 = K14e_0
  H42_0 = K24e_0 + 2.5 * K14e_0

  # Collisionality-dependent Hmn (Eqs. 30b, 30e)
  temp1 = nu_e_star_eff * ftrap_d**3 * (1.0 + ftrap_d**6)
  H11 = (
      H11_0
      / (
          1.0
          + a_coeffs[:, 0, 0] * nu_e_star_eff_sqrt
          + b_coeffs[:, 0, 0] * nu_e_star_eff
      )
      - d_coeffs[:, 0, 0] * temp1 / (1.0 + c_coeffs[:, 0, 0] * temp1) * FPS
  )
  H12 = (
      H12_0
      / (
          1.0
          + a_coeffs[:, 0, 1] * nu_e_star_eff_sqrt
          + b_coeffs[:, 0, 1] * nu_e_star_eff
      )
      - d_coeffs[:, 0, 1] * temp1 / (1.0 + c_coeffs[:, 0, 1] * temp1) * FPS
  )
  H22 = (
      H22_0
      / (
          1.0
          + a_coeffs[:, 1, 1] * nu_e_star_eff_sqrt
          + b_coeffs[:, 1, 1] * nu_e_star_eff
      )
      - d_coeffs[:, 1, 1] * temp1 / (1.0 + c_coeffs[:, 1, 1] * temp1) * FPS
  )

  temp2 = 1.0 / (1.0 + nu_e_star_eff**2 * ftrap**12)
  temp4 = nu_e_star_eff * ftrap**3 * (1.0 + 0.8 * ftrap**3)
  H41 = (
      H41_0
      / (
          1.0
          + a_coeffs[:, 0, 0] * nu_e_star_eff_sqrt
          + b_coeffs[:, 0, 0] * nu_e_star_eff
      )
      - d_coeffs[:, 0, 0] * temp4 / (1.0 + c_coeffs[:, 0, 0] * temp4) * FPS4
  ) * temp2
  H42 = (
      H42_0
      / (
          1.0
          + a_coeffs[:, 0, 1] * nu_e_star_eff_sqrt
          + b_coeffs[:, 0, 1] * nu_e_star_eff
      )
      - d_coeffs[:, 0, 1] * temp4 / (1.0 + c_coeffs[:, 0, 1] * temp4) * FPS4
  ) * temp2

  # Electron Kmn matrix (Eq. 30a, 30d)
  Kmn_e = jnp.zeros((ftrap.shape[0], 4, 4))
  Kmn_e = Kmn_e.at[:, 0, 0].set(H11)
  Kmn_e = Kmn_e.at[:, 0, 1].set(H12 - 2.5 * H11)
  Kmn_e = Kmn_e.at[:, 1, 0].set(Kmn_e[:, 0, 1])
  Kmn_e = Kmn_e.at[:, 1, 1].set(H22 - 5.0 * H12 + 6.25 * H11)
  Kmn_e = Kmn_e.at[:, 0, 3].set(H41)
  Kmn_e = Kmn_e.at[:, 3, 0].set(Kmn_e[:, 0, 3])
  Kmn_e = Kmn_e.at[:, 1, 3].set(H42 - 2.5 * H41)
  Kmn_e = Kmn_e.at[:, 3, 1].set(Kmn_e[:, 1, 3])

  # Supplement K matrix with "bootstrap terms" needed for Ware pinch from the
  # Sauter model (PoP 1999)
  Kmn_e = Kmn_e.at[:, 0, 2].set(
      -formulas.calculate_L31(ftrap, nu_e_star, Z_eff)
  )
  Kmn_e = Kmn_e.at[:, 2, 0].set(Kmn_e[:, 0, 2])
  Kmn_e = Kmn_e.at[:, 1, 2].set(
      -formulas.calculate_L32(ftrap, nu_e_star, Z_eff)
  )
  Kmn_e = Kmn_e.at[:, 2, 1].set(Kmn_e[:, 1, 2])

  # Ion Kmn matrix
  # alpha coefficient, Eq. (25)
  alpha = (
      -(0.62 + 1.5 * alpha_I)
      / (0.53 + alpha_I)
      * ((1.0 - ftrap) / (1.0 - 0.22 * ftrap - 0.19 * ftrap**2))
  )

  # Eq. 24d
  F22_i_ftrapd = (1.0 - 0.55) * (
      1.0 + 1.54 * alpha_I
  ) * ftrap_d + ftrap_d**2 * (0.75 + ftrap_d * (-0.7 + 0.5 * ftrap_d)) * (
      1.0 + 2.92 * alpha_I
  )

  # K11, for completeness. Bottom of page 1230.
  K11_i = 1.0 / (0.11 + 1.7 * ftrap - 1.25 * ftrap**2 + 0.44 * ftrap**3) - 1.0

  # K22_i, Eq. 30h
  mu_i_star_eff = nu_i_star_eff * (1.0 + 1.54 * alpha_I)
  Hp = 1.0 + 1.33 * alpha_I * (1.0 + 0.60 * alpha_I) / (1.0 + 1.79 * alpha_I)
  temp4_i = mu_i_star_eff * ftrap_d**3 * (1.0 + ftrap_d**6)
  K22_i = (
      -F22_i_ftrapd / (1 + a2 * jnp.sqrt(mu_i_star_eff) + b2 * mu_i_star_eff)
      - d2 * temp4_i / (1.0 + c2 * temp4_i) * Hp * FPS
  )

  Kmn_i = jnp.zeros((ftrap.shape[0], 2, 2))
  Kmn_i = Kmn_i.at[:, 0, 0].set(K11_i)
  Kmn_i = Kmn_i.at[:, 1, 1].set(K22_i)
  Kmn_i = Kmn_i.at[:, 0, 1].set(-alpha)  # Eq. 25
  Kmn_i = Kmn_i.at[:, 1, 0].set(alpha)

  return Kmn_e, Kmn_i


def _Fmn_X(
    X: array_typing.FloatVectorFace, Z_eff: array_typing.FloatVectorFace
) -> array_typing.Array:
  """Calculates the F_mn matrix from Eq. (24) of Angioni & Sauter 2000."""
  F11 = X + X * (0.9 + X * (-1.9 + X * (1.6 - 0.6 * X))) / (Z_eff + 0.5)
  F12 = X + X * (0.6 + X * (-0.95 + X * (0.3 + 0.05 * X))) / (Z_eff + 0.5)
  F22 = X + X * (-0.11 + X * (0.08 + 0.03 * X)) / (Z_eff + 0.5)
  # Transpose such that the leading dimension is the same as the input arrays.
  return jnp.array([[F11, F12], [F12, F22]]).transpose(2, 0, 1)


def _coeffs_appendix_B(
    Z_eff: array_typing.Array,
) -> tuple[
    array_typing.Array,
    array_typing.Array,
    array_typing.Array,
    array_typing.Array,
    float,
    float,
    float,
    float,
]:
  """Calculates coefficients from Appendix B of Angioni & Sauter 2000."""
  a11 = (1.0 + 3.0 * Z_eff) / (0.77 + 1.22 * Z_eff)
  a12 = (0.72 + 0.42 * Z_eff) / (1.0 + 0.5 * Z_eff)
  a22 = 0.46 * jnp.ones_like(a11)
  # Transpose such that the leading dimension is the same as the input arrays.
  a_coeffs = jnp.array([[a11, a12], [a12, a22]]).transpose(2, 0, 1)

  b11 = (1.0 + 1.1 * Z_eff) / (1.37 * Z_eff)
  b12 = (1.0 + Z_eff) / (2.99 * Z_eff)
  b22 = Z_eff / (-3.0 + 5.32 * Z_eff)
  b_coeffs = jnp.array([[b11, b12], [b12, b22]]).transpose(2, 0, 1)

  c11 = (0.1 + 0.34 * Z_eff) / (1.65 * Z_eff)
  c12 = (0.27 + 0.4 * Z_eff) / (1.0 + 3.0 * Z_eff)
  c22 = (0.22 + 0.55 * Z_eff) / (-1.0 + 7.0 * Z_eff)
  c_coeffs = jnp.array([[c11, c12], [c12, c22]]).transpose(2, 0, 1)

  d11 = 0.23 * Z_eff / (-1.0 + 3.85 * Z_eff)
  d12 = (0.22 + 0.38 * Z_eff) / (1.0 + 6.1 * Z_eff)
  d22 = (0.25 + 0.05 * Z_eff) / (1.0 + 0.82 * Z_eff)

  d_coeffs = jnp.array([[d11, d12], [d12, d22]]).transpose(2, 0, 1)
  a2, b2, c2, d2 = 1.03, 0.31, 0.22, 0.175
  return a_coeffs, b_coeffs, c_coeffs, d_coeffs, a2, b2, c2, d2


def _calculate_Lmn(
    Kmn_e: array_typing.Array,
    Kmn_i: array_typing.Array,
    geo: geometry_lib.Geometry,
    core_profiles: state.CoreProfiles,
    epsilon: array_typing.FloatVectorFace,
    nu_e_star: array_typing.FloatVectorFace,
    nu_i_star: array_typing.FloatVectorFace,
) -> tuple[array_typing.Array, array_typing.Array]:
  """Calculates the dimensional transport matrices Lmn."""
  # Normalization factors from Eqs. 16, 20, 21
  consts = constants.CONSTANTS
  thermal_velocity_e = jnp.sqrt(
      2 * core_profiles.T_e.face_value() * consts.keV2J / consts.me
  )
  collision_time_e = (core_profiles.q_face * geo.R_major) / (
      nu_e_star * epsilon**1.5 * thermal_velocity_e + consts.eps
  )
  thermal_velocity_i = jnp.sqrt(
      2
      * core_profiles.T_i.face_value()
      * consts.keV2J
      / (core_profiles.A_i * consts.m_amu)
  )
  collision_time_i = (core_profiles.q_face * geo.R_major) / (
      nu_i_star * epsilon**1.5 * thermal_velocity_i + consts.eps
  )

  r_larmor_e = consts.me * thermal_velocity_e / consts.qe
  r_larmor_i = (
      consts.m_amu
      * core_profiles.A_i
      * thermal_velocity_i
      / (consts.qe * core_profiles.Z_i_face)
  )

  dpsi_dr = core_profiles.psi.face_grad() / geo.rho_b

  Ld = (
      core_profiles.n_e.face_value()
      * r_larmor_e**2
      / collision_time_e
      * dpsi_dr**2
  )
  Ldi = (
      core_profiles.n_i.face_value()
      * r_larmor_i**2
      / collision_time_i
      * dpsi_dr**2
  )
  Lb = geo.F_face * core_profiles.n_e.face_value()
  Lbi = geo.F_face * core_profiles.n_i.face_value()

  Lsi = (
      core_profiles.n_i.face_value()
      * (consts.qe * core_profiles.Z_i_face) ** 2
      * collision_time_i
      * geo.B_0**2
      / (
          consts.m_amu
          * core_profiles.A_i
          * core_profiles.T_i.face_value()
          * consts.keV2J
      )
  )

  # Calculate electron matrix (Eq. 20)
  Lmn_e = jnp.zeros((nu_e_star.shape[0], 4, 4))

  Lmn_e = Lmn_e.at[:, 0, 0].set(Kmn_e[:, 0, 0] * Ld * geo.gm4_face * geo.B_0**2)
  Lmn_e = Lmn_e.at[:, 0, 1].set(Kmn_e[:, 0, 1] * Ld * geo.gm4_face * geo.B_0**2)
  Lmn_e = Lmn_e.at[:, 0, 2].set(Kmn_e[:, 0, 2] * Lb)
  Lmn_e = Lmn_e.at[:, 0, 3].set(Kmn_e[:, 0, 3] * Ld / geo.gm5_face * geo.B_0**2)

  Lmn_e = Lmn_e.at[:, 1, 0].set(Lmn_e[:, 0, 1])
  Lmn_e = Lmn_e.at[:, 1, 1].set(Kmn_e[:, 1, 1] * Ld * geo.gm4_face * geo.B_0**2)
  Lmn_e = Lmn_e.at[:, 1, 2].set(Kmn_e[:, 1, 2] * Lb)
  Lmn_e = Lmn_e.at[:, 1, 3].set(Kmn_e[:, 1, 3] * Ld / geo.gm5_face * geo.B_0**2)

  Lmn_e = Lmn_e.at[:, 2, 0].set(Lmn_e[:, 0, 2])
  Lmn_e = Lmn_e.at[:, 2, 1].set(Lmn_e[:, 1, 2])
  # Lmn_e[:, 2 ,2] is not used
  Lmn_e = Lmn_e.at[:, 2, 3].set(Kmn_e[:, 2, 3] * Lb)

  Lmn_e = Lmn_e.at[:, 3, 0].set(Lmn_e[:, 0, 3])
  Lmn_e = Lmn_e.at[:, 3, 1].set(Lmn_e[:, 1, 3])
  Lmn_e = Lmn_e.at[:, 3, 2].set(Lmn_e[:, 2, 3])
  Lmn_e = Lmn_e.at[:, 3, 3].set(Kmn_e[:, 3, 3] * Ld / geo.gm5_face * geo.B_0**2)

  # Calculate ion matrix
  Lmn_i = jnp.zeros((nu_i_star.shape[0], 2, 2))
  Lmn_i = Lmn_i.at[:, 0, 0].set(
      Kmn_i[:, 0, 0] * Lsi * geo.gm5_face / geo.B_0**2
  )
  Lmn_i = Lmn_i.at[:, 0, 1].set(Kmn_i[:, 0, 1] * Lbi)
  Lmn_i = Lmn_i.at[:, 1, 0].set(-Lmn_i[:, 1, 0])
  Lmn_i = Lmn_i.at[:, 1, 1].set(
      Kmn_i[:, 1, 1] * Ldi * geo.gm4_face * geo.B_0**2
  )

  return Lmn_e, Lmn_i



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common formulas used in neoclassical models."""

import jax.numpy as jnp
from torax._src import array_typing
from torax._src import constants
from torax._src.geometry import geometry as geometry_lib

# pylint: disable=invalid-name


def calculate_f_trap(
    geo: geometry_lib.Geometry,
) -> array_typing.FloatVectorFace:
  """Calculates the effective trapped particle fraction.

  From O. Sauter, Fusion Engineering and Design 112 (2016) 633-645. Eqs 33+34.

  Args:
    geo: The magnetic geometry.

  Returns:
    The effective trapped particle fraction.
  """

  epsilon_effective = (
      0.67
      * (1.0 - 1.4 * jnp.abs(geo.delta_face) * geo.delta_face)
      * geo.epsilon_face
  )
  aa = (1.0 - geo.epsilon_face) / (1.0 + geo.epsilon_face)
  return 1.0 - jnp.sqrt(aa) * (1.0 - epsilon_effective) / (
      1.0 + 2.0 * jnp.sqrt(epsilon_effective)
  )


def calculate_L31(
    f_trap: array_typing.FloatVectorFace,
    nu_e_star: array_typing.FloatVectorFace,
    Z_eff: array_typing.FloatVectorFace,
) -> array_typing.FloatVectorFace:
  """Calculates the L31 Sauter coefficient: Sauter PoP 1999 Eqs. 14a + 14b."""
  denom = (
      1.0
      + (1 - 0.1 * f_trap) * jnp.sqrt(nu_e_star)
      + 0.5 * (1.0 - f_trap) * nu_e_star / Z_eff
  )
  ft31 = f_trap / denom
  term_0 = (1 + 1.4 / (Z_eff + 1)) * ft31
  term_1 = -1.9 / (Z_eff + 1) * ft31**2
  term_2 = 0.3 / (Z_eff + 1) * ft31**3
  term_3 = 0.2 / (Z_eff + 1) * ft31**4
  return term_0 + term_1 + term_2 + term_3


def calculate_L32(
    f_trap: array_typing.FloatVectorFace,
    nu_e_star: array_typing.FloatVectorFace,
    Z_eff: array_typing.FloatVectorFace,
) -> array_typing.FloatVectorFace:
  """Calculates the L32 Sauter coefficient: Sauter PoP 1999 Eqs. 15a-e."""
  ft32ee = f_trap / (
      1
      + 0.26 * (1 - f_trap) * jnp.sqrt(nu_e_star)
      + 0.18 * (1 - 0.37 * f_trap) * nu_e_star / jnp.sqrt(Z_eff)
  )
  ft32ei = f_trap / (
      1
      + (1 + 0.6 * f_trap) * jnp.sqrt(nu_e_star)
      + 0.85 * (1 - 0.37 * f_trap) * nu_e_star * (1 + Z_eff)
  )

  F32ee = (
      (0.05 + 0.62 * Z_eff)
      / (Z_eff * (1 + 0.44 * Z_eff))
      * (ft32ee - ft32ee**4)
      + 1
      / (1 + 0.22 * Z_eff)
      * (ft32ee**2 - ft32ee**4 - 1.2 * (ft32ee**3 - ft32ee**4))
      + 1.2 / (1 + 0.5 * Z_eff) * ft32ee**4
  )

  F32ei = (
      -(0.56 + 1.93 * Z_eff)
      / (Z_eff * (1 + 0.44 * Z_eff))
      * (ft32ei - ft32ei**4)
      + 4.95
      / (1 + 2.48 * Z_eff)
      * (ft32ei**2 - ft32ei**4 - 0.55 * (ft32ei**3 - ft32ei**4))
      - 1.2 / (1 + 0.5 * Z_eff) * ft32ei**4
  )
  return F32ee + F32ei


# TODO(b/428166775): currently we have two very similar implementations for
# nu_e_star. We should refactor this to have a single one in physics/collisions
def calculate_nu_e_star(
    q: array_typing.FloatVectorFace,
    geo: geometry_lib.Geometry,
    n_e: array_typing.FloatVectorFace,
    T_e: array_typing.FloatVectorFace,
    Z_eff: array_typing.FloatVectorFace,
    log_lambda_ei: array_typing.FloatVectorFace,
) -> array_typing.FloatVectorFace:
  """Calculates the electron collisionality, nu_e_star.

  This is the electron collisionality, defined as the ratio of the electron
  collision frequency to the bounce frequency. From Sauter PoP 1999 Eq. (18b).

  Args:
    q: Safety factor.
    geo: The geometry of the torus.
    n_e: Electron density [m^-3].
    T_e: Electron temperature [keV]. Converted to eV in the formula.
    Z_eff: Effective charge.
    log_lambda_ei: Electron-ion Coulomb logarithm.

  Returns:
    The electron collisionality.
  """
  return (
      6.921e-18
      * q
      * geo.R_major
      * n_e
      * Z_eff
      * log_lambda_ei
      / (
          ((T_e * 1e3) ** 2)
          * (geo.epsilon_face + constants.CONSTANTS.eps) ** 1.5
      )
  )


def calculate_nu_i_star(
    q: array_typing.FloatVectorFace,
    geo: geometry_lib.Geometry,
    n_i: array_typing.FloatVectorFace,
    T_i: array_typing.FloatVectorFace,
    Z_eff: array_typing.FloatVectorFace,
    log_lambda_ii: array_typing.FloatVectorFace,
) -> array_typing.FloatVectorFace:
  """Calculates the ion collisionality, nu_i_star.

  This is the ion collisionality, defined as the ratio of the ion
  collision frequency to the bounce frequency. From Sauter PoP 1999 Eq. (18c).

  Args:
    q: Safety factor.
    geo: The geometry of the torus.
    n_i: Ion density.
    T_i: Ion temperature.
    Z_eff: Effective charge.
    log_lambda_ii: Ion-ion Coulomb logarithm.

  Returns:
    The ion collisionality.
  """
  return (
      4.9e-18
      * q
      * geo.R_major
      * n_i
      * Z_eff**4
      * log_lambda_ii
      / (
          ((T_i * 1e3) ** 2)
          * (geo.epsilon_face + constants.CONSTANTS.eps) ** 1.5
      )
  )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Collisional radiative model for charge states and cooling rates.

From A. A. Mavrin (2017):
Radiative Cooling Rates for Low-Z Impurities in Non-coronal Equilibrium State.
J Fusion Energ (2017) 36:161-172
DOI 10.1007/s10894-017-0136-z
"""

import enum
import jax
from jax import numpy as jnp
from torax._src import array_typing
from torax._src.edge import mavrin_2017_charge_states_data
from torax._src.edge import mavrin_2017_cooling_rate_data

# pylint: disable=invalid-name

_NE_TAU_CORONAL_LIMIT = 1e19


class MavrinVariable(enum.StrEnum):
  Z = 'Z'
  LZ = 'LZ'


def calculate_mavrin_2017(
    T_e: array_typing.FloatVector,
    ne_tau: array_typing.FloatScalar,
    ion_symbol: str,
    variable: MavrinVariable,
) -> array_typing.FloatVector:
  """Calculates the average charge state of an impurity based on a polynomial fit.

  Polynomial fit range is ~1 eV-15 keV, which is within the typical
  bounds of tokamak edge plasmas. For safety, inputs are clipped to avoid
  polynomial extrapolation outside the precise per-species ranges.

  Args:
    T_e: Electron temperature [keV].
    ne_tau: The non-coronal parameter, being the product of electron density and
      impurity residence time [m^-3 s]. High values of ne_tau correspond to the
      coronal limit.
    ion_symbol: Species to calculate average charge state for.
    variable: The variable to calculate, either 'Z' (charge states) or 'LZ'
      (radiative cooling rates).

  Returns:
    Either average charge states or cooling rate, depending on the variable.
  """

  match variable:
    case MavrinVariable.Z:
      cr_module = mavrin_2017_charge_states_data
    case MavrinVariable.LZ:
      cr_module = mavrin_2017_cooling_rate_data
    case _:
      allowed_variables = ', '.join([v.name for v in MavrinVariable])
      raise ValueError(
          f'Invalid fit variable: {variable}. Allowed variables are:'
          f' {allowed_variables}'
      )

  if ion_symbol not in cr_module.COEFFS.keys():
    raise ValueError(
        f'Invalid ion symbol: {ion_symbol}. Allowed symbols are:'
        f' {cr_module.COEFFS.keys()}'
    )

  # Mavrin 2017 formulas are constructed for [eV] temperature input
  T_e_ev = T_e * 1e3

  # Avoid extrapolating fitted polynomial out of bounds.
  min_temp = cr_module.MIN_TEMPERATURES[ion_symbol]
  max_temp = cr_module.MAX_TEMPERATURES[ion_symbol]
  T_e_ev = jnp.clip(T_e_ev, min_temp, max_temp)
  # Residence parameter capped at 10^19, which is the coronal limit.
  ne_tau = jnp.clip(ne_tau, a_max=_NE_TAU_CORONAL_LIMIT)

  # Gather coefficients for each temperature
  interval_indices = jnp.searchsorted(
      cr_module.TEMPERATURE_INTERVALS[ion_symbol], T_e_ev
  )
  coeffs_in_range = jnp.take(
      cr_module.COEFFS[ion_symbol], interval_indices, axis=1
  )

  X = jnp.log10(T_e_ev)
  Y = jnp.log10(ne_tau / _NE_TAU_CORONAL_LIMIT)

  # 2D polynomial from Mavrin 2017, Eq. 8
  log10_variable = (
      coeffs_in_range[0]
      + coeffs_in_range[1] * X
      + coeffs_in_range[2] * Y
      + coeffs_in_range[3] * X**2
      + coeffs_in_range[4] * X * Y
      + coeffs_in_range[5] * Y**2
      + coeffs_in_range[6] * X**3
      + coeffs_in_range[7] * X**2 * Y
      + coeffs_in_range[8] * X * Y**2
      + coeffs_in_range[9] * Y**3
  )

  return 10**log10_variable


def calculate_L_INT(
    start_temp: array_typing.FloatScalar,
    stop_temp: array_typing.FloatScalar,
    ne_tau: array_typing.FloatScalar,
    ion_symbol: str,
    resolution: int = 100,
) -> array_typing.FloatScalar:
  """Calculates the integral of Lz * sqrt(Te) from start_temp to stop_temp.

  This is the L_INT term from the extended Lengyel model.

  Equation 34 from T. Body, et al., Nucl. Fusion 65 (2025) 086002

  Args:
    start_temp: The starting electron temperature for the integration [keV].
    stop_temp: The stopping electron temperature for the integration [keV].
    ne_tau: The non-coronal parameter [m^-3 s].
    ion_symbol: The impurity species.
    resolution: The number of points to use in the numerical integration.

  Returns:
    The integrated cooling rate [eV^1.5 * m^3 * W].
  """
  # Convert start and stop temperatures to eV for the integration grid
  start_temp_eV = start_temp * 1e3
  stop_temp_eV = stop_temp * 1e3

  # Create a logarithmically spaced grid of electron temperatures in eV
  electron_temp_ev = jnp.logspace(
      jnp.log10(start_temp_eV), jnp.log10(stop_temp_eV), num=resolution
  )

  # Convert the grid back to keV to call the existing Mavrin model function
  electron_temp = electron_temp_ev / 1e3

  # Calculate Lz at each temperature point on the grid
  Lz_values = calculate_mavrin_2017(
      T_e=electron_temp,
      ne_tau=ne_tau,
      ion_symbol=ion_symbol,
      variable=MavrinVariable.LZ,
  )

  # The integrand is Lz * sqrt(Te)
  Lz_sqrt_Te = Lz_values * jnp.sqrt(electron_temp_ev)

  # Perform trapezoidal integration
  Lint = jax.scipy.integrate.trapezoid(y=Lz_sqrt_Te, x=electron_temp_ev)

  return Lint



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The CellVariable class.

A jax_utils.jax_dataclass used to represent variables on meshes for the 1D fvm
solver.
Naming conventions and API are similar to those developed in the FiPy fvm solver
[https://www.ctcms.nist.gov/fipy/]
"""
import dataclasses

import chex
import jax
from jax import numpy as jnp
import jaxtyping as jt
from torax._src import array_typing
import typing_extensions


def _zero() -> array_typing.FloatScalar:
  """Returns a scalar zero as a jax Array."""
  return jnp.zeros(())


@chex.dataclass(frozen=True)
class CellVariable:
  """A variable representing values of the cells along the radius.

  Attributes:
    value: A jax.Array containing the value of this variable at each cell.
    dr: Distance between cell centers.
    left_face_constraint: An optional jax scalar specifying the value of the
      leftmost face. Defaults to None, signifying no constraint. The user can
      modify this field at any time, but when face_grad is called exactly one of
      left_face_constraint and left_face_grad_constraint must be None.
    left_face_grad_constraint: An optional jax scalar specifying the (otherwise
      underdetermined) value of the leftmost face. See left_face_constraint.
    right_face_constraint: Analogous to left_face_constraint but for the right
      face, see left_face_constraint.
    right_face_grad_constraint: A jax scalar specifying the undetermined value
      of the gradient on the rightmost face variable.
  """

  # t* means match 0 or more leading time dimensions.
  value: jt.Float[chex.Array, 't* cell']
  dr: jt.Float[chex.Array, 't*']
  left_face_constraint: jt.Float[chex.Array, 't*'] | None = None
  right_face_constraint: jt.Float[chex.Array, 't*'] | None = None
  left_face_grad_constraint: jt.Float[chex.Array, 't*'] | None = (
      dataclasses.field(default_factory=_zero)
  )
  right_face_grad_constraint: jt.Float[chex.Array, 't*'] | None = (
      dataclasses.field(default_factory=_zero)
  )
  # Can't make the above default values be jax zeros because that would be a
  # call to jax before absl.app.run

  def __post_init__(self):
    """Check that the CellVariable is valid.

    How is `sanity_check` different from `__post_init__`?
    - `sanity_check` is exposed to the client directly, so the client can
    explicitly check sanity without violating privacy conventions. This is
    useful for checking objects that were created e.g. using jax tree
    transformations.
    - `sanity_check` is guaranteed not to change the object, while
    `__post_init__` could in principle make changes.
    """
    # Automatically check dtypes of all numeric fields

    for field in dataclasses.fields(self):
      value = getattr(self, field.name)
      name = field.name
      if isinstance(value, jax.Array):
        if value.dtype != jnp.float64 and jax.config.read('jax_enable_x64'):
          raise TypeError(
              f'Expected dtype float64, got dtype {value.dtype} for `{name}`'
          )
        if value.dtype != jnp.float32 and not jax.config.read('jax_enable_x64'):
          raise TypeError(
              f'Expected dtype float32, got dtype {value.dtype} for `{name}`'
          )
    left_and = (
        self.left_face_constraint is not None
        and self.left_face_grad_constraint is not None
    )
    left_or = (
        self.left_face_constraint is not None
        or self.left_face_grad_constraint is not None
    )
    if left_and or not left_or:
      raise ValueError(
          'Exactly one of left_face_constraint and '
          'left_face_grad_constraint must be set.'
      )
    right_and = (
        self.right_face_constraint is not None
        and self.right_face_grad_constraint is not None
    )
    right_or = (
        self.right_face_constraint is not None
        or self.right_face_grad_constraint is not None
    )
    if right_and or not right_or:
      raise ValueError(
          'Exactly one of right_face_constraint and '
          'right_face_grad_constraint must be set.'
      )

  def _assert_unbatched(self):
    if len(self.value.shape) != 1:
      raise AssertionError(
          'CellVariable must be unbatched, but has `value` shape '
          f'{self.value.shape}. Consider using vmap to batch the function call.'
      )
    if self.dr.shape:
      raise AssertionError(
          'CellVariable must be unbatched, but has `dr` shape '
          f'{self.dr.shape}. Consider using vmap to batch the function call.'
      )

  def face_grad(
      self, x: jt.Float[chex.Array, 'cell'] | None = None
  ) -> jt.Float[chex.Array, 'face']:
    """Returns the gradient of this value with respect to the faces.

    Implemented using forward differencing of cells. Leftmost and rightmost
    gradient entries are determined by user specify constraints, see
    CellVariable class docstring.

    Args:
      x: (optional) coordinates over which differentiation is carried out

    Returns:
      A jax.Array of shape (num_faces,) containing the gradient.
    """
    self._assert_unbatched()
    if x is None:
      forward_difference = jnp.diff(self.value) / self.dr
    else:
      forward_difference = jnp.diff(self.value) / jnp.diff(x)

    def constrained_grad(
        face: jax.Array | None,
        grad: jax.Array | None,
        cell: jax.Array,
        right: bool,
    ) -> jax.Array:
      """Calculates the constrained gradient entry for an outer face."""

      if face is not None:
        if grad is not None:
          raise ValueError(
              'Cannot constraint both the value and gradient of '
              'a face variable.'
          )
        if x is None:
          dx = self.dr
        else:
          dx = x[-1] - x[-2] if right else x[1] - x[0]
        sign = -1 if right else 1
        return sign * (cell - face) / (0.5 * dx)
      else:
        if grad is None:
          raise ValueError('Must specify one of value or gradient.')
        return grad

    left_grad = constrained_grad(
        self.left_face_constraint,
        self.left_face_grad_constraint,
        self.value[0],
        right=False,
    )
    right_grad = constrained_grad(
        self.right_face_constraint,
        self.right_face_grad_constraint,
        self.value[-1],
        right=True,
    )

    left = jnp.expand_dims(left_grad, axis=0)
    right = jnp.expand_dims(right_grad, axis=0)
    return jnp.concatenate([left, forward_difference, right])

  def _left_face_value(self) -> jt.Float[chex.Array, '#t']:
    """Calculates the value of the leftmost face."""
    if self.left_face_constraint is not None:
      value = self.left_face_constraint
      # Boundary value has one fewer dim than cell value, expand to concat with.
      value = jnp.expand_dims(value, axis=-1)
    else:
      # When there is no constraint, leftmost face equals
      # leftmost cell
      value = self.value[..., 0:1]
    return value

  def _right_face_value(self) -> jt.Float[chex.Array, '#t']:
    """Calculates the value of the rightmost face."""
    if self.right_face_constraint is not None:
      value = self.right_face_constraint
      # Boundary value has one fewer dim than cell value, expand to concat with.
      value = jnp.expand_dims(value, axis=-1)
    else:
      # Maintain right_face consistent with right_face_grad_constraint
      value = (
          self.value[..., -1:]
          + jnp.expand_dims(self.right_face_grad_constraint, axis=-1)
          * jnp.expand_dims(self.dr, axis=-1)
          / 2
      )
    return value

  def face_value(self) -> jt.Float[jax.Array, 't* face']:
    """Calculates values of this variable on the face grid."""
    inner = (self.value[..., :-1] + self.value[..., 1:]) / 2.0
    return jnp.concatenate(
        [self._left_face_value(), inner, self._right_face_value()], axis=-1
    )

  def grad(self) -> jt.Float[jax.Array, 't* face']:
    """Returns the gradient of this variable wrt cell centers."""
    face = self.face_value()
    return jnp.diff(face) / jnp.expand_dims(self.dr, axis=-1)

  def __str__(self) -> str:
    output_string = f'CellVariable(value={self.value}'
    if self.left_face_constraint is not None:
      output_string += f', left_face_constraint={self.left_face_constraint}'
    if self.right_face_constraint is not None:
      output_string += f', right_face_constraint={self.right_face_constraint}'
    if self.left_face_grad_constraint is not None:
      output_string += (
          f', left_face_grad_constraint={self.left_face_grad_constraint}'
      )
    if self.right_face_grad_constraint is not None:
      output_string += (
          f', right_face_grad_constraint={self.right_face_grad_constraint}'
      )
    output_string += ')'
    return output_string

  def cell_plus_boundaries(self) -> jt.Float[jax.Array, 't* cell+2']:
    """Returns the value of this variable plus left and right boundaries."""
    right_value = self._right_face_value()
    left_value = self._left_face_value()
    return jnp.concatenate(
        [left_value, self.value, right_value],
        axis=-1,
    )

  def __eq__(self, other: typing_extensions.Self) -> bool:
    try:
      chex.assert_trees_all_equal(self, other)
      return True
    except AssertionError:
      return False
